from collections.abc import Callable, MutableMapping
from dataclasses import dataclass, field
from importlib import import_module
from io import StringIO
import json
import os
from pathlib import Path
import re
from typing import TYPE_CHECKING, Any, ClassVar, Literal, TypeVar
import warnings

from tarina.tools import nest_dict_update, safe_eval

from .model import Proxy, config_model_dump, config_model_schema, config_model_validate
from .models.default import BasicConfModel as BasicConfModel
from .models.default import field as model_field
from .util import GetattrDict

try:
    from ruamel.yaml import YAML
except ImportError:
    YAML = None

if TYPE_CHECKING:
    from ..plugin import Plugin

EXPR_CONTEXT_PAT = re.compile(r"['\"]?\$\{\{\s?(?P<expr>[^}\s]+)\s?\}\}['\"]?")
T = TypeVar("T")
T_M = TypeVar("T_M", bound=MutableMapping)


class WebsocketsInfo(BasicConfModel):
    """Satori Server WebSocket Configuration"""

    type: Literal["websocket", "websockets", "ws"]
    host: str = model_field(default="localhost", description="WebSocket server host")
    port: int = model_field(default=5140, description="WebSocket server port")
    path: str = model_field(default="", description="WebSocket server endpoint path")
    token: str | None = model_field(default=None, description="Authentication token for the WebSocket server")
    timeout: float | None = model_field(default=None, description="Connection timeout in seconds")


class WebhookInfo(BasicConfModel):
    """Satori Server Webhook Configuration"""

    type: Literal["webhook", "wh", "http"]
    host: str = model_field(default="127.0.0.1", description="Webhook self-server host")
    port: int = model_field(default=8080, description="Webhook self-server port")
    path: str = model_field(default="v1/events", description="Webhook self-server endpoint path")
    token: str | None = model_field(default=None, description="Authentication token for the webhook")
    server_host: str = model_field(default="localhost", description="Target server host")
    server_port: int = model_field(default=5140, description="Target server port")
    server_path: str = model_field(default="", description="Target server endpoint path")
    timeout: float | None = model_field(default=None, description="Connection timeout in seconds")


class LogSaveInfo(BasicConfModel):
    """Configuration for saving logs to a file"""

    rotation: str = model_field(default="00:00", description="Log rotation time, e.g., '00:00' for daily rotation")
    compression: str | None = model_field(default=None, description="Compression format for log saving, e.g., 'zip'")
    colorize: bool = model_field(default=True, description="Whether to colorize the log output")


class LogInfo(BasicConfModel):
    """Configuration for the application logs"""

    level: int | str = model_field(default="INFO", description="Log level for the application")
    ignores: list[str] = model_field(default_factory=list, description="Log ignores for the application")
    save: LogSaveInfo | bool | None = model_field(
        default=None,
        description="Log saving configuration, if None or False, logs will not be saved",
    )
    rich_error: bool = model_field(default=False, description="Whether enable rich traceback for exceptions")
    short_level: bool = model_field(default=False, description="Whether use short log level names")


class BasicConfig(BasicConfModel):
    """Basic configuration for the Entari application"""

    network: list[WebsocketsInfo | WebhookInfo] = model_field(default_factory=list, description="Network configuration")
    ignore_self_message: bool = model_field(default=True, description="Whether ignore self-send message event")
    skip_req_missing: bool = model_field(
        default=False, description="Whether skip Event Handler if requirement is missing"
    )
    log: LogInfo = model_field(default_factory=LogInfo, description="Log configuration")
    log_level: int | str | None = model_field(default=None, description="[Deprecated] Log level for the application")
    log_ignores: list[str] | None = model_field(
        default=None, description="[Deprecated] Log ignores for the application"
    )
    prefix: list[str] = model_field(default_factory=list, description="Command prefix for the application")
    cmd_count: int = model_field(default=4096, description="Command count limit for the application")
    external_dirs: list[str] = model_field(default_factory=list, description="External directories to look for plugins")
    schema: bool = model_field(
        default=False, description="Whether generate JSON schema for the configuration (after application start)"
    )

    def __post_init__(self):
        if self.log_level is not None:
            self.log.level = self.log_level
        if self.log_ignores is not None:
            self.log.ignores = self.log_ignores
        if self.prefix.count(""):
            self.prefix = [p for p in self.prefix if p]
            self.prefix.append("")


_loaders: dict[str, Callable[[str], dict]] = {}
_dumpers: dict[str, Callable[[dict, int], str]] = {}


@dataclass
class EntariConfig:
    path: Path
    basic: BasicConfig = field(default_factory=BasicConfig, init=False)
    plugin: dict[str, dict] = field(default_factory=dict, init=False)
    prelude_plugin: list[str] = field(default_factory=list, init=False)
    plugin_extra_files: list[str] = field(default_factory=list, init=False)
    save_flag: bool = field(default=False)
    _origin_data: dict[str, Any] = field(init=False)
    _env_replaced: dict[int, str] = field(default_factory=dict, init=False)

    instance: ClassVar["EntariConfig"]

    def loader(self, path: Path):
        if not path.exists():
            return {}
        end = path.suffix.split(".")[-1]
        if end in _loaders:
            ctx = {"env": GetattrDict(os.environ)}

            with path.open("r", encoding="utf-8") as f:
                lines = f.readlines()
            for i, line in enumerate(lines):

                def handle(m: re.Match[str]):
                    self._env_replaced[i] = line
                    expr = m.group("expr")
                    return safe_eval(expr, ctx)

                lines[i] = EXPR_CONTEXT_PAT.sub(handle, line)
            text = "".join(lines)
            return _loaders[end](text)

        raise ValueError(f"Unsupported file format: {path.suffix}")

    def dumper(self, path: Path, save_path: Path, data: dict, indent: int):
        if not path.exists():
            return
        origin = self.loader(path)
        if "entari" in origin:
            origin["entari"] = data
        else:
            origin = data
        end = save_path.suffix.split(".")[-1]
        if end in _dumpers:
            ans = _dumpers[end](origin, indent)
            if self._env_replaced:
                lines = ans.splitlines(keepends=True)
                for i, line in self._env_replaced.items():
                    lines[i] = line
                ans = "".join(lines)
            with save_path.open("w", encoding="utf-8") as f:
                f.write(ans)
            return
        raise ValueError(f"Unsupported file format: {save_path.suffix}")

    def __post_init__(self):
        self.__class__.instance = self
        self.reload()

    @property
    def data(self) -> dict[str, Any]:
        return self._origin_data

    @property
    def prelude_plugin_names(self) -> list[str]:
        return [name for name in self.plugin_names if name in self.prelude_plugin]

    @property
    def plugin_names(self) -> list[str]:
        slots = [
            (name, self.plugin[name].get("$priority", 16))
            for name in self.plugin
            if not name.startswith("$") and not self.plugin[name].get("$optional", False)
        ]
        slots.sort(key=lambda x: x[1])
        return [name for name, _ in slots]

    def reload(self):
        if self.save_flag:
            self.save_flag = False
            return False
        data = self.loader(self.path)
        if "entari" in data:
            data = data["entari"]
        self.basic = config_model_validate(BasicConfig, data.get("basic", {}))
        self._origin_data = data
        self.plugin = data.get("plugins", {})
        self.plugin_extra_files: list[str] = self.plugin.get("$files", [])  # type: ignore
        self.prelude_plugin = self.plugin.get("$prelude", [])  # type: ignore
        for key in list(self.plugin.keys()):
            if key.startswith("$"):
                continue
            value = self.plugin.pop(key)
            if key.startswith("~"):
                key = key[1:]
                value["$disable"] = True
            elif key.startswith("?"):
                key = key[1:]
                value["$optional"] = True
            self.plugin[key] = value
        for file in self.plugin_extra_files:
            path = Path(file)
            if not path.exists():
                raise FileNotFoundError(file)
            if path.is_dir():
                for _path in path.iterdir():
                    if not _path.is_file():
                        continue
                    self.plugin[_path.stem] = self.loader(_path)
            else:
                self.plugin[path.stem] = self.loader(path)
        return True

    @staticmethod
    def _clean(value: T_M) -> T_M:
        value.pop("$path", None)
        value.pop("$static", None)
        return value

    def dump(self, indent: int = 2):
        basic = self._origin_data.get("basic", {})
        if "log" not in basic and ("log_level" in basic or "log_ignores" in basic):
            basic["log"] = {}
            if "log_level" in basic:
                basic["log"]["level"] = basic.pop("log_level")
            if "log_ignores" in basic:
                basic["log"]["ignores"] = basic.pop("log_ignores")

        if self.plugin_extra_files:
            for file in self.plugin_extra_files:
                path = Path(file)
                if path.is_file():
                    self.dumper(path, path, self._clean(self.plugin.pop(path.stem)), indent)
                else:
                    for _path in path.iterdir():
                        if _path.is_file():
                            self.dumper(_path, _path, self._clean(self.plugin.pop(_path.stem)), indent)
        for key in list(self.plugin.keys()):
            if key.startswith("$"):
                continue
            value = self.plugin.pop(key)
            if "$disable" in value:
                key = f"~{key}" if value["$disable"] else key
                value.pop("$disable", None)
            if "$optional" in value:
                key = f"?{key}" if value["$optional"] else key
                value.pop("$optional", None)
            self.plugin[key] = self._clean(value)
        return self._origin_data

    def save(self, path: str | os.PathLike[str] | None = None, indent: int = 2):
        self.save_flag = True
        self.dumper(self.path, Path(path or self.path), self.dump(indent), indent)

    @classmethod
    def load(cls, path: str | os.PathLike[str] | None = None) -> "EntariConfig":
        try:
            import dotenv

            dotenv.load_dotenv()
        except ImportError:
            dotenv = None  # noqa
            pass
        if not path:
            if "ENTARI_CONFIG_FILE" in os.environ:
                _path = Path(os.environ["ENTARI_CONFIG_FILE"])
            elif (Path.cwd() / ".entari.json").exists():
                _path = Path.cwd() / ".entari.json"
            else:
                _path = Path.cwd() / "entari.yml"
        else:
            _path = Path(path)
        if "ENTARI_CONFIG_EXTENSION" in os.environ:
            ext_mods = os.environ["ENTARI_CONFIG_EXTENSION"].split(";")
            for ext_mod in ext_mods:
                if not ext_mod:
                    continue
                ext_mod = ext_mod.replace("::", "arclet.entari.config.format.")
                try:
                    import_module(ext_mod)
                except ImportError as e:
                    warnings.warn(f"Failed to load config extension '{ext_mod}': {e}", ImportWarning)
        if not _path.exists():
            return cls(_path)
        if not _path.is_file():
            raise ValueError(f"{_path} is not a file")
        return cls(_path)

    def bind(self, plugin: str, obj: T) -> T:
        """
        Bind the plugin object to the config, allowing the config to be updated when the object changes.
        """
        if plugin not in self.plugin:
            raise KeyError(f"Plugin Config {plugin} not found in config")

        def updater(target):
            nest_dict_update(self.plugin[plugin], config_model_dump(target))
            self.save()

        ans = Proxy(obj, lambda target=obj: updater(target))
        return ans  # type: ignore

    def generate_schema(self, plugins: list["Plugin"]):
        # fmt: off
        schemas = {
            "basic": config_model_schema(BasicConfig, ref_root="/properties/basic/"),
            "plugins": {
                "type": "object", "description": "Plugin configurations", "properties": {"$prelude": {"type": "array", "items": {"type": "string", "description": "Plugin name"}, "description": "List of prelude plugins to load", "default": [], "uniqueItems": True}, "$files": {"type": "array", "items": {"type": "string", "description": "File path"}, "description": "List of configuration files to load", "default": [], "uniqueItems": True}, **{plug._config_key: ((config_model_schema(plug.metadata.config, ref_root=f"/properties/plugins/properties/{plug._config_key}/") if plug.metadata.config is not None else {"type": "object", "description": f"{plug.metadata.description or plug.metadata.name}; no configuration required", "additionalProperties": True}) if plug.metadata else {"type": "object", "description": "No configuration required", "additionalProperties": True}) for plug in plugins}}  # noqa: E501
            }
        }
        with open(f"{self.path.stem}.schema.json", "w", encoding="utf-8") as f:
            json.dump({"$schema": "https://json-schema.org/draft/2020-12/schema", "type": "object", "properties": schemas, "additionalProperties": False, "required": ["basic"]}, f, indent=2, ensure_ascii=False)  # noqa: E501
        # fmt: on


load_config = EntariConfig.load


def register_loader(*ext: str):
    """Register a loader for a specific file extension."""

    def decorator(func: Callable[[str], dict]):
        for e in ext:
            _loaders[e] = func
        return func

    return decorator


def register_dumper(*ext: str):
    """Register a dumper for a specific file extension."""

    def decorator(func: Callable[[dict, int], str]):
        for e in ext:
            _dumpers[e] = func
        return func

    return decorator


@register_loader("json")
def json_loader(text: str) -> dict:
    return json.loads(text)


@register_dumper("json")
def json_dumper(origin: dict, indent: int):
    return json.dumps(origin, indent=indent, ensure_ascii=False)


@register_loader("yaml", "yml")
def yaml_loader(text: str) -> dict:
    if YAML is None:
        raise RuntimeError("yaml is not installed. Please install with `arclet-entari[yaml]`")
    yaml = YAML()
    yaml.preserve_quotes = True
    yaml.indent(mapping=2, sequence=4, offset=2)
    return yaml.load(StringIO(text))


@register_dumper("yaml", "yml")
def yaml_dumper(origin: dict, indent: int):
    if YAML is None:
        raise RuntimeError("yaml is not installed. Please install with `arclet-entari[yaml]`")
    yaml = YAML()
    yaml.preserve_quotes = True
    yaml.indent(mapping=indent, sequence=indent + 2, offset=indent)
    sio = StringIO()
    yaml.dump(origin, sio)
    return sio.getvalue()



from typing import Any

from arclet.entari.config.model import ConfigModelAction

try:
    from pydantic import BaseModel, Field, TypeAdapter, json_schema

    class PydanticConfigAction(ConfigModelAction[BaseModel]):
        """Pydantic Config Model Action"""

        @classmethod
        def load(cls, data: dict[str, Any], t: type):
            return TypeAdapter(t).validate_python(data)

        @classmethod
        def dump(cls, obj):
            return obj.model_dump()

        @classmethod
        def keys(cls, obj) -> list[str]:
            return list(obj.model_fields.keys())

        @classmethod
        def schema(cls, t: type[BaseModel], ref_root: str = "/"):  # type: ignore
            return json_schema.model_json_schema(t, ref_template=f"#{ref_root}$defs/{{model}}")

except ImportError:
    BaseModel = None
    Field = None


def __getattr__(name):
    if name == "BaseModel":
        if BaseModel is None:
            raise ImportError("Please install `pydantic` first. Install with `pip install arclet-entari[model]`")
        return BaseModel
    if name == "Field":
        if Field is None:
            raise ImportError("Please install `pydantic` first. Install with `pip install arclet-entari[model]`")
        return Field
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = ["BaseModel", "Field"]



import inspect
from typing import Any, Literal, Union, get_args

from arclet.alconna import Alconna, Arparma, Duplication, Empty, output_manager
from arclet.alconna.builtin import generate_duplication
from arclet.alconna.exceptions import SpecialOptionTriggered
from arclet.letoderea import STOP, Contexts, Param, Propagator, Provider, post
from arclet.letoderea.exceptions import ProviderUnsatisfied
from arclet.letoderea.provider import ProviderFactory
from nepattern.util import CUnionType
from satori.element import Text
from tarina.generic import get_origin

from ..config import EntariConfig
from ..event.base import Reply
from ..event.command import CommandOutput, CommandParse, CommandReceive
from ..message import MessageChain
from ..session import Session
from .model import CommandResult, Match, Query


def _remove_config_prefix(message: MessageChain):
    if not (command_prefix := EntariConfig.instance.basic.prefix):
        return message
    if message and isinstance(message[0], Text):
        text = message[0].text.lstrip()  # type: ignore
        for prefix in command_prefix:
            if not prefix:
                return message
            if text.startswith(prefix):
                message[0] = Text(text[len(prefix) :])
                return message
    return MessageChain()


class MessageJudges(Propagator):
    def __init__(self, need_reply_me: bool, need_notice_me: bool, use_config_prefix: bool):
        self.need_reply_me = need_reply_me
        self.need_notice_me = need_notice_me
        self.use_config_prefix = use_config_prefix

    async def judge(self, ctx: Contexts, message: MessageChain, is_reply_me: bool = False, is_notice_me: bool = False):
        message = message.fork()
        if self.need_reply_me and not is_reply_me:
            return STOP
        if self.need_notice_me and not is_notice_me:
            return STOP
        if self.use_config_prefix and not (message := _remove_config_prefix(message)):
            return STOP
        if "$message_content" in ctx:
            return {"$message_content": message}
        return {"$message": message}

    def compose(self):
        yield self.judge, True, 60


class AlconnaSuppiler(Propagator):
    cmd: Alconna

    def __init__(self, cmd: Alconna):
        self.cmd = cmd

    async def before_supply(self, message: MessageChain, session: Session | None = None, reply: Reply | None = None):
        if session:
            recv = await post(ev := CommandReceive(session, self.cmd, message, reply))
            message = recv.value if recv else ev.content
        return {"_message": message}

    async def supply(self, ctx: Contexts):
        try:
            _message: MessageChain = ctx.pop("_message")
        except KeyError:
            raise ProviderUnsatisfied("_message") from None
        with output_manager.capture(self.cmd.name) as cap:
            output_manager.set_action(lambda x: x, self.cmd.name)
            try:
                _res = self.cmd.parse(_message)
            except Exception as e:
                _res = Arparma(self.cmd._hash, _message, False, error_info=e)
            may_help_text: str | None = cap.get("output", None)
        return {"_result": CommandResult(self.cmd, _res, may_help_text)}

    async def after_supply(self, ctx: Contexts, session: Session | None = None):
        try:
            alc_result: CommandResult = ctx.pop("_result")
        except KeyError:
            raise ProviderUnsatisfied("_result") from None
        _res = alc_result.result
        if session:
            pres = await post(ev := CommandParse(session, self.cmd, _res))
            _res = ev.result
            if pres:
                if isinstance(pres.value, Arparma):
                    _res = pres.value
                elif pres.value is False:
                    return STOP
        if _res.matched:
            return {"alc_result": CommandResult(self.cmd, _res, alc_result.output)}
        if alc_result.output:
            if session:
                _t = str(_res.error_info) if isinstance(_res.error_info, SpecialOptionTriggered) else "error"
                ores = await post(ev := CommandOutput(session, self.cmd, _t, alc_result.output))
                msg = MessageChain(ev.content)
                if ores:
                    if ores.value is False:
                        return STOP
                    elif isinstance(ores.value, str | MessageChain):
                        msg = MessageChain(ores.value)
                await session.send(msg)
                return STOP
            return {"alc_result": alc_result}
        return STOP

    def compose(self):
        yield self.before_supply, True, 65
        yield self.supply, True, 70
        yield self.after_supply, True, 75


class AlconnaProvider(Provider[Any]):
    def __init__(self, type_: str, extra: dict | None = None):
        super().__init__()
        self.type = type_
        self.extra = extra or {}

    async def __call__(self, context: Contexts):
        if "alc_result" not in context:
            if self.type == "args":
                return
            raise ProviderUnsatisfied("alc_result")
        result: CommandResult = context["alc_result"]
        if self.type == "result":
            return result
        if self.type == "arparma":
            return result.result
        if self.type == "alconna":
            return result.source
        if self.type == "default_duplication":
            return generate_duplication(result.source)(result.result)
        if self.type == "duplication":
            return self.extra["duplication"](result.result)
        if self.type == "match":
            target = result.result.all_matched_args.get(self.extra["name"], Empty)
            return Match(target, target != Empty)
        if self.type == "query":
            q = Query(self.extra["query"].path, self.extra["query"].result)
            res = result.result.query(q.path, Empty)
            q.available = res != Empty
            if q.available:
                q.result = res
            elif self.extra["query"].result != Empty:
                q.available = True
            return q
        if self.extra["name"] in result.result.all_matched_args:
            return result.result.all_matched_args[self.extra["name"]]


_seminal = type("_seminal", (object,), {})


class Assign(Propagator):
    def __init__(self, path: str, value: Any = _seminal, or_not: bool = False):
        self.path = path
        self.value = value
        self.or_not = or_not

    async def check(self, alc_result: CommandResult):
        if self.value == _seminal:
            if self.path == "$main" or self.or_not:
                if not alc_result.result.components:
                    return
                return STOP
            if alc_result.result.query(self.path, "\1") == "\1":
                return STOP
        else:
            if alc_result.result.query(self.path) != self.value:
                return STOP
            if self.or_not and alc_result.result.query(self.path) == Empty:
                return
            return STOP

    def compose(self):
        yield self.check, True, 80


class AlconnaProviderFactory(ProviderFactory):
    def validate(self, param: Param):
        annotation = get_origin(param.annotation)
        if annotation in (Union, CUnionType, Literal):
            annotation = get_origin(get_args(param.annotation)[0])
        if annotation is CommandResult:
            return AlconnaProvider("result")
        if annotation is Arparma:
            return AlconnaProvider("arparma")
        if annotation is Alconna:
            return AlconnaProvider("alconna")
        if annotation is Duplication:
            return AlconnaProvider("default_duplication")
        if inspect.isclass(annotation) and issubclass(annotation, Duplication):
            return AlconnaProvider("duplication", {"duplication": param.annotation})
        if annotation is Match:
            return AlconnaProvider("match", {"name": param.name})
        if isinstance(param.default, Query):
            return AlconnaProvider("query", {"query": param.default})
        return AlconnaProvider("args", {"name": param.name, "anno": param.annotation})



from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime
from typing import Any, ClassVar, TypeVar, overload

from arclet.letoderea import Contexts, Param, Provider, define
from satori import ArgvInteraction, ButtonInteraction, Channel
from satori import Event as OriginEvent
from satori import EventType, Guild, Member, Role, User
from satori.client import Account
from satori.element import At, Author, Quote, Text, select
from satori.model import Login, MessageObject
from tarina import gen_subclass

from ..message import MessageChain

T = TypeVar("T")
D = TypeVar("D")


@dataclass
class Reply:
    quote: Quote
    origin: MessageObject


def _is_reply_me(reply: Reply, account: Account):
    if reply.origin.user:
        return reply.origin.user.id == account.self_id
    if authors := select(reply.quote, Author):
        return any(author.id == account.self_id for author in authors)
    return False


def _is_notice_me(message: MessageChain, account: Account):
    if message and isinstance(message[0], At):
        at: At = message[0]  # type: ignore
        if at.id and at.id == account.self_id:
            return True
    return False


def _remove_notice_me(message: MessageChain, account: Account):
    message = message.copy()
    message.pop(0)
    if _is_notice_me(message, account):
        message.pop(0)
    if message and isinstance(message[0], Text):
        text = message[0].text.lstrip()  # type: ignore
        if not text:
            message.pop(0)
        else:
            message[0] = Text(text)
    return message


class Attr:
    def __init__(self, key: str | None = None, cls: Callable[..., T] | None = None, internal: bool = False):
        self.key = key or ""
        self.cls_ = cls
        self.internal = internal

    def __set_name__(self, owner: type["SatoriEvent"], name: str):
        self.key = self.key or name
        if name not in ("id", "timestamp"):
            owner._attrs.add(name)

    def __get__(self, instance: "SatoriEvent", owner: type["SatoriEvent"]):
        if self.internal and instance._origin._data:
            val = instance._origin._data.get(self.key)
        else:
            val = getattr(instance._origin, self.key, None)
        if self.cls_ and val is not None:
            return self.cls_(val)
        return val

    def __set__(self, instance: "SatoriEvent", value):
        raise AttributeError("can't set attribute")


@overload
def attr(*, internal: bool = False) -> Any: ...


@overload
def attr(key: str, /, *, internal: bool = False) -> Any: ...


@overload
def attr(cls: Callable[..., T], /, *, internal: bool = False) -> T: ...


@overload
def attr(key: str, cls: Callable[..., T], /, *, internal: bool = False) -> T: ...


def attr(*args, internal: bool = False) -> Any:
    if not args:
        return Attr(internal=internal)
    key = args[0] if isinstance(args[0], str) else None
    cls = args[1] if len(args) == 2 else (args[0] if not isinstance(args[0], str) else None)
    return Attr(key, cls, internal)


class SatoriEvent:
    type: ClassVar[str]
    _attrs: ClassVar[set[str]] = set()
    _origin: OriginEvent
    account: Account

    sn: int = attr()
    timestamp: datetime = attr()
    login: Login = attr()
    argv: ArgvInteraction | None = attr()
    button: ButtonInteraction | None = attr()
    channel: Channel | None = attr()
    guild: Guild | None = attr()
    member: Member | None = attr()
    message: MessageObject | None = attr()
    operator: User | None = attr()
    role: Role | None = attr()
    user: User | None = attr()

    def __init_subclass__(cls, **kwargs):
        cls._attrs = set()

    def __init__(self, account: Account, origin: OriginEvent):
        self.account = account
        self._origin = origin

    async def gather(self, context: Contexts):
        context["$account"] = self.account
        context["$origin_event"] = self._origin

        for name in self.__class__._attrs:
            value = getattr(self, name)
            if value is not None:
                context["$message_origin" if name == "message" else f"${name}"] = value

    class TimeProvider(Provider[datetime]):
        def validate(self, param: Param):
            return super().validate(param) and param.name == "event_time"

        async def __call__(self, context: Contexts):
            if "$event" in context:
                return context["$event"].timestamp

    class OperatorProvider(Provider[User]):
        priority = 10

        def validate(self, param: Param):
            return param.name == "operator" and super().validate(param)

        async def __call__(self, context: Contexts):
            if "$operator" in context:
                return context["$operator"]
            if "$origin_event" not in context:
                return
            return context["$origin_event"].operator

    class UserProvider(Provider[User]):
        async def __call__(self, context: Contexts):
            if "$user" in context:
                return context["$user"]
            if "$origin_event" not in context:
                return
            return context["$origin_event"].user

    class MessageProvider(Provider[MessageObject]):
        async def __call__(self, context: Contexts):
            if "$message_origin" in context:
                return context["$message_origin"]
            if "$origin_event" not in context:
                return
            return context["$origin_event"].message

    class ChannelProvider(Provider[Channel]):
        async def __call__(self, context: Contexts):
            if "$channel" in context:
                return context["$channel"]
            if "$origin_event" not in context:
                return
            return context["$origin_event"].channel

    class GuildProvider(Provider[Guild]):
        async def __call__(self, context: Contexts):
            if "$guild" in context:
                return context["$guild"]
            if "$origin_event" not in context:
                return
            return context["$origin_event"].guild

    class MemberProvider(Provider[Member]):
        async def __call__(self, context: Contexts):
            if "$member" in context:
                return context["$member"]
            if "$origin_event" not in context:
                return
            return context["$origin_event"].member

    class RoleProvider(Provider[Role]):
        async def __call__(self, context: Contexts):
            if "$role" in context:
                return context["$role"]
            if "$origin_event" not in context:
                return
            return context["$origin_event"].role

    class LoginProvider(Provider[Login]):
        async def __call__(self, context: Contexts):
            if "$login" in context:
                return context["$login"]
            if "$origin_event" not in context:
                return
            return context["$origin_event"].login

    def __repr__(self):
        return f"<{self.__class__.__name__.removesuffix('Event')}{self._origin!r}>"


class NoticeEvent(SatoriEvent):
    pass


class FriendEvent(NoticeEvent):
    user: User = attr()


class FriendRequestEvent(FriendEvent):
    type = EventType.FRIEND_REQUEST

    message: MessageObject = attr()


class GuildEvent(NoticeEvent):
    guild: Guild = attr()

    async def gather(self, context: Contexts):
        await super().gather(context)
        context["$guild"] = self.guild


class GuildAddedEvent(GuildEvent):
    type = EventType.GUILD_ADDED


class GuildRemovedEvent(GuildEvent):
    type = EventType.GUILD_REMOVED


class GuildRequestEvent(GuildEvent):
    type = EventType.GUILD_REQUEST

    message: MessageObject = attr()


class GuildUpdatedEvent(GuildEvent):
    type = EventType.GUILD_UPDATED


class GuildMemberEvent(GuildEvent):
    user: User = attr()


class GuildMemberAddedEvent(GuildMemberEvent):
    type = EventType.GUILD_MEMBER_ADDED


class GuildMemberRemovedEvent(GuildMemberEvent):
    type = EventType.GUILD_MEMBER_REMOVED


class GuildMemberRequestEvent(GuildMemberEvent):
    type = EventType.GUILD_MEMBER_REQUEST

    message: MessageObject = attr()


class GuildMemberUpdatedEvent(GuildMemberEvent):
    type = EventType.GUILD_MEMBER_UPDATED


class GuildRoleEvent(GuildEvent):
    role: Role = attr()


class GuildRoleCreatedEvent(GuildRoleEvent):
    type = EventType.GUILD_ROLE_CREATED


class GuildRoleDeletedEvent(GuildRoleEvent):
    type = EventType.GUILD_ROLE_DELETED


class GuildRoleUpdatedEvent(GuildRoleEvent):
    type = EventType.GUILD_ROLE_UPDATED


class LoginEvent(NoticeEvent):
    pass


class LoginAddedEvent(LoginEvent):
    type = EventType.LOGIN_ADDED


class LoginRemovedEvent(LoginEvent):
    type = EventType.LOGIN_REMOVED


class LoginUpdatedEvent(LoginEvent):
    type = EventType.LOGIN_UPDATED


class MessageContentProvider(Provider[MessageChain]):
    priority = 30

    async def __call__(self, context: Contexts):
        return context.get("$message_content")


class ReplyProvider(Provider[Reply]):
    async def __call__(self, context: Contexts):
        return context.get("$message_reply")


class MessageEvent(SatoriEvent):
    channel: Channel = attr()
    user: User = attr()
    message: MessageObject = attr()

    content: MessageChain
    quote: Quote | None = None

    providers = [MessageContentProvider, ReplyProvider]

    def __init__(self, account: Account, origin: OriginEvent):
        super().__init__(account, origin)
        self.content = MessageChain(self.message.message)
        if self.content.has(Quote):
            self.quote = self.content.get(Quote, 1)[0]
            self.content = self.content.exclude(Quote)

    async def gather(self, context: Contexts):
        await super().gather(context)
        reply = None
        if self.quote and self.quote.id:
            if self.quote.children:
                mo = MessageObject.from_elements(self.quote.id, self.quote.children)
            else:
                mo = await self.account.protocol.message_get(self.channel.id, self.quote.id)
            reply = context["$message_reply"] = Reply(self.quote, mo)
        if not reply:
            is_reply_me = False
        else:
            is_reply_me = _is_reply_me(reply, self.account)
        context["is_reply_me"] = is_reply_me
        if is_reply_me and self.content and isinstance(self.content[0], Text):
            text = self.content[0].text.lstrip()
            if not text:
                self.content.pop(0)
            else:
                self.content[0] = Text(text)
        is_notice_me = context["is_notice_me"] = _is_notice_me(self.content, self.account)
        if is_notice_me:
            self.content = _remove_notice_me(self.content, self.account)
        context["$message_content"] = self.content


class MessageCreatedEvent(MessageEvent):
    type = EventType.MESSAGE_CREATED


class MessageDeletedEvent(MessageEvent):
    type = EventType.MESSAGE_DELETED


class MessageUpdatedEvent(MessageEvent):
    type = EventType.MESSAGE_UPDATED


class ReactionEvent(NoticeEvent, MessageEvent):
    pass


class ReactionAddedEvent(ReactionEvent):
    type = EventType.REACTION_ADDED


class ReactionRemovedEvent(ReactionEvent):
    type = EventType.REACTION_REMOVED


class InternalEvent(SatoriEvent):
    type = EventType.INTERNAL


class InteractionEvent(NoticeEvent):
    pass


class InteractionButtonEvent(InteractionEvent):
    type = EventType.INTERACTION_BUTTON

    button: ButtonInteraction = attr()

    class ButtonProvider(Provider[ButtonInteraction]):
        async def __call__(self, context: Contexts):
            return context.get("button")


class InteractionCommandEvent(InteractionEvent):
    type = EventType.INTERACTION_COMMAND


class InteractionCommandArgvEvent(InteractionCommandEvent):
    argv: ArgvInteraction = attr()

    class ArgvProvider(Provider[ArgvInteraction]):
        async def __call__(self, context: Contexts):
            return context.get("argv")


class InteractionCommandMessageEvent(InteractionCommandEvent, MessageEvent):
    pass


MAPPING: dict[str, type[SatoriEvent]] = {}

for cls in gen_subclass(SatoriEvent):
    if hasattr(cls, "type"):
        typ = cls.type.value if isinstance(cls.type, EventType) else cls.type
        MAPPING[typ] = cls
        define(cls, name=typ)


INTERNAL_ADDITIONAL_HANDLERS: list[Callable[[str, str, dict[str, Any]], type[SatoriEvent] | None]] = []


def event_parse(account: Account, event: OriginEvent):
    constructor_cls = MAPPING.get(event.type)
    if (constructor_cls is None or event.type == EventType.INTERNAL) and event._type and event._data:
        found = next((h(event.type, event._type, event._data) for h in INTERNAL_ADDITIONAL_HANDLERS), None)
        if found is not None:
            constructor_cls = found
            define(constructor_cls, name=constructor_cls.type)
    if constructor_cls is None:
        raise NotImplementedError(f"Unsupported event type: {event.type}")
    return constructor_cls(account, event)


def register_internal_event(func: Callable[[str, str, dict[str, Any]], type[SatoriEvent] | None]):
    INTERNAL_ADDITIONAL_HANDLERS.append(func)



from typing import Any, Literal

from arclet.letoderea import Result, make_event


@make_event(name="entari.event/config_reload")
class ConfigReload:
    scope: Literal["basic", "plugin"]
    key: str
    value: Any
    old: Any = None

    def __post_init__(self):
        self.key = self.key.lstrip("~")

    def check_result(self, value) -> Result[bool] | None:
        if isinstance(value, bool):
            return Result(value)



# This file is @generated by tarina.lang CLI tool
# It is not intended for manual editing.

from tarina.lang.model import LangItem, LangModel


class EntariPluginInspectScene:
    name: LangItem = LangItem("entari::plugin_inspect", "scene.name")
    direct: LangItem = LangItem("entari::plugin_inspect", "scene.direct")
    text: LangItem = LangItem("entari::plugin_inspect", "scene.text")
    voice: LangItem = LangItem("entari::plugin_inspect", "scene.voice")
    category: LangItem = LangItem("entari::plugin_inspect", "scene.category")


class EntariPluginInspect:
    description: LangItem = LangItem("entari::plugin_inspect", "description")
    platform: LangItem = LangItem("entari::plugin_inspect", "platform")
    self: LangItem = LangItem("entari::plugin_inspect", "self")
    scene = EntariPluginInspectScene
    user: LangItem = LangItem("entari::plugin_inspect", "user")
    channel: LangItem = LangItem("entari::plugin_inspect", "channel")
    guild: LangItem = LangItem("entari::plugin_inspect", "guild")
    invalid: LangItem = LangItem("entari::plugin_inspect", "invalid")


class Lang(LangModel):
    entari_plugin_inspect = EntariPluginInspect



from arclet.letoderea import Depends as Depends
from arclet.letoderea import bind as bind
from arclet.letoderea import depends as depends
from arclet.letoderea import deref as deref
from arclet.letoderea import enter_if as enter_if
from arclet.letoderea import param as param
from arclet.letoderea import propagate as propagate
from arclet.letoderea import use as use
from arclet.letoderea.core import ExceptionEvent as ExceptionEvent
from satori import ArgvInteraction as ArgvInteraction
from satori import At as At
from satori import Audio as Audio
from satori import Author as Author
from satori import Bold as Bold
from satori import Button as Button
from satori import ButtonInteraction as ButtonInteraction
from satori import Channel as Channel
from satori import ChannelType as ChannelType
from satori import Code as Code
from satori import E as E
from satori import Element as Element
from satori import Event as Event
from satori import File as File
from satori import Guild as Guild
from satori import Image as Image
from satori import Italic as Italic
from satori import Link as Link
from satori import Login as Login
from satori import LoginStatus as LoginStatus
from satori import Member as Member
from satori import Message as Message
from satori import MessageObject as MessageObject
from satori import Quote as Quote
from satori import Role as Role
from satori import Sharp as Sharp
from satori import Spoiler as Spoiler
from satori import Strikethrough as Strikethrough
from satori import Subscript as Subscript
from satori import Superscript as Superscript
from satori import Text as Text
from satori import Underline as Underline
from satori import User as User
from satori import Video as Video
from satori import transform as transform
from satori.client import Account as Account
from satori.client.config import WebhookInfo as WebhookInfo
from satori.client.config import WebsocketsInfo as WebsocketsInfo

from . import command as command
from . import scheduler as scheduler
from .config import BasicConfModel as BasicConfModel
from .config import load_config as load_config
from .core import Entari as Entari
from .event import BaseEvent as BaseEvent
from .event import attr as attr
from .event import register_internal_event as register_internal_event
from .event.base import MessageCreatedEvent as MessageCreatedEvent
from .event.base import MessageEvent as MessageEvent
from .event.base import Reply as Reply
from .event.config import ConfigReload as ConfigReload
from .event.lifespan import AccountUpdate as AccountUpdate
from .event.lifespan import Cleanup as Cleanup
from .event.lifespan import Ready as Ready
from .event.lifespan import Startup as Startup
from .event.send import SendRequest as SendRequest
from .event.send import SendResponse as SendResponse
from .filter import filter_ as filter_
from .localdata import local_data as local_data
from .message import MessageChain as MessageChain
from .plugin import Plugin as Plugin
from .plugin import PluginMetadata as PluginMetadata
from .plugin import add_service as add_service
from .plugin import collect_disposes as collect_disposes
from .plugin import declare_static as declare_static
from .plugin import disable_plugin as disable_plugin
from .plugin import enable_plugin as enable_plugin
from .plugin import keeping as keeping
from .plugin import listen as listen
from .plugin import load_plugin as load_plugin
from .plugin import load_plugins as load_plugins
from .plugin import metadata as metadata
from .plugin import package as package
from .plugin import plugin_config as plugin_config
from .plugin import requires as requires
from .plugin import unload_plugin as unload_plugin
from .plugin.model import inject as inject
from .session import Session as Session

Param = param
WS = WebsocketsInfo
WH = WebhookInfo

__version__ = "0.16.8"



from __future__ import annotations

import asyncio
from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from arclet.letoderea import es
from launart import Launart, Service
from launart.status import Phase

from ..event.lifespan import Cleanup, Ready, Startup
from ..event.plugin import PluginUnloaded
from ..logger import log

if TYPE_CHECKING:
    from .model import KeepingVariable, Plugin, RootlessPlugin


class ServiceWaiters:
    def __init__(self):
        self._futures: dict[str, asyncio.Future] = {}

    def assign(self, service_id: str):
        if service_id not in self._futures:
            self._futures[service_id] = asyncio.Future()
        self._futures[service_id].set_result(True)

    def clear(self, service_id: str):
        if service_id in self._futures:
            future = self._futures.pop(service_id)
            if not future.done():
                future.set_result(True)

    async def wait_for(self, service_id: str):
        if service_id not in self._futures:
            self._futures[service_id] = asyncio.Future()
        await self._futures[service_id]


class PluginManagerService(Service):
    id = "entari.plugin.manager"

    plugins: dict[str, Plugin]
    _keep_values: dict[str, dict[str, KeepingVariable]]
    referents: dict[str, set[str]]
    references: dict[str, set[str]]
    _direct_plugins: set[str]
    _unloaded: set[str]
    _subplugined: dict[str, str]
    _apply: dict[str, tuple[Callable[[dict[str, Any]], RootlessPlugin], bool]]

    def __init__(self):
        super().__init__()
        self.plugins = {}
        self._direct_plugins = set()
        self._keep_values = {}
        self.referents = {}
        self.references = {}
        self._unloaded = set()
        self._subplugined = {}
        self._apply = {}
        self.service_waiter = ServiceWaiters()

    @property
    def required(self) -> set[str]:
        return {"entari.service"}

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "cleanup", "blocking"}

    async def launch(self, manager: Launart):

        for plug in self.plugins.values():
            for serv in plug._services.values():
                manager.add_component(serv)
                self.service_waiter.assign(serv.id)
            if plug.config.get("$disable", False):
                plug.disable()

        async with self.stage("preparing"):
            es.publish(Startup())
        async with self.stage("blocking"):
            for plug in self.plugins.values():
                if not plug._apply:
                    continue
                plug.exec_apply()
                if plug.config.get("$disable", False):
                    plug.disable()
            es.publish(Ready())
            await manager.status.wait_for_sigexit()
        async with self.stage("cleanup"):
            es.publish(Cleanup())
            ids = [k for k in self.plugins.keys() if k not in self._subplugined]
            for plug_id in reversed(ids):
                plug = self.plugins[plug_id]
                try:
                    await es.publish(PluginUnloaded(plug.id))
                    plug.dispose(is_cleanup=True)
                except Exception as e:
                    log.plugin.error(f"failed to dispose plugin <y>{plug.id}</y> caused by {e!r}")
                    self.plugins.pop(plug_id, None)
            for values in self._keep_values.values():
                for value in values.values():
                    value.dispose()
                values.clear()
            self._keep_values.clear()


plugin_service = PluginManagerService()



from collections.abc import Callable
from dataclasses import MISSING, Field, asdict, dataclass
from dataclasses import field as _field
from dataclasses import fields, is_dataclass
from datetime import date, datetime
from inspect import Signature
from pathlib import Path
from typing import Any, ForwardRef, TypeVar, get_args, overload
from typing_extensions import dataclass_transform

from tarina import generic_isinstance
from tarina.generic import get_origin, origin_is_union
from tarina.signature import merge_cls_and_parent_ns

from ..dc_schema import SchemaGenerator
from ..model import ConfigModelAction
from ..util import store_field_description

_available_dc_attrs = set(Signature.from_callable(dataclass).parameters.keys())
_available_field_attrs = set(Signature.from_callable(Field).parameters.keys())

C = TypeVar("C")
_T = TypeVar("_T")

# fmt: off


@overload
def field(*, default: _T, init: bool = True, repr: bool = True, description: str | None = None, hash: bool | None = None, compare: bool = True, metadata: dict[str, Any] | None = None) -> _T: ...  # noqa: E501


@overload
def field(*, default_factory: Callable[[], _T], init: bool = True, repr: bool = True, description: str | None = None, hash: bool | None = None, compare: bool = True, metadata: dict[str, Any] | None = None) -> _T: ...  # noqa: E501


@overload
def field(*, init: bool = True, repr: bool = True, description: str | None = None, hash: bool | None = None, compare: bool = True, metadata: dict[str, Any] | None = None) -> Any: ...  # noqa: E501


def field(*, default=MISSING, default_factory=MISSING, init=True, repr=True, description=None, hash=None, compare=True, metadata=None):  # noqa: E501  # type: ignore
    if default is not MISSING and default_factory is not MISSING:
        raise ValueError("cannot specify both default and default_factory")
    if metadata is None and description is not None:
        metadata = {"description": description}
    data = {"default": default, "default_factory": default_factory, "init": init, "repr": repr, "hash": hash, "compare": compare, "metadata": metadata, "kw_only": True}  # noqa: E501
    return Field(**{k: v for k, v in data.items() if k in _available_field_attrs})

# fmt: on


@dataclass_transform(kw_only_default=True, field_specifiers=(field, _field))
class BasicConfModel:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        dataclass(**({k: v for k, v in (kwargs | {"kw_only": True}).items() if k in _available_dc_attrs}))(cls)
        store_field_description(cls, cls.__dataclass_fields__)  # type: ignore


def _resolve_type(field_type: Any, types_namespace: dict[str, Any]) -> Any:
    """解析字段类型，处理字符串类型引用"""
    if isinstance(field_type, str):
        return ForwardRef(field_type, is_argument=True, is_class=True)._evaluate(
            types_namespace, types_namespace, recursive_guard=frozenset()
        )
    return field_type


def _validate_union_type(value: Any, union_type: Any) -> tuple[bool, Any]:
    """验证 Union 类型，返回 (是否成功, 验证后的值)"""
    args = get_args(union_type)
    non_none_args = tuple(tp for tp in args if tp is not type(None))

    # 如果值为 None 且 Union 包含 NoneType，则直接返回
    if value is None and type(None) in args:
        return True, None

    for arg_type in non_none_args:
        try:
            if is_dataclass(arg_type):
                if isinstance(value, dict):
                    validated_value = _nested_validate(value, arg_type)
                    return True, validated_value
                elif generic_isinstance(value, arg_type):
                    return True, value
            elif generic_isinstance(value, arg_type):
                return True, value
            else:
                validated_value = _validate_complex_type(value, arg_type)
                if validated_value is not None:
                    return True, validated_value
        except (TypeError, ValueError, AttributeError):
            continue

    return False, value


def _validate_complex_type(value: Any, tp: Any) -> Any:
    """验证复杂类型（list, dict, set, tuple 等）"""
    orig_tp = get_origin(tp)

    if orig_tp is list and isinstance(value, list):
        args = get_args(tp)
        if args and is_dataclass(args[0]):
            return [_nested_validate(d, args[0]) for d in value]
        elif args:
            return [_validate_single_value(d, args[0]) for d in value]
        return value

    elif orig_tp is set and isinstance(value, set | list):
        args = get_args(tp)
        if args and is_dataclass(args[0]):
            return {_nested_validate(d, args[0]) for d in value}
        elif args:
            return {_validate_single_value(d, args[0]) for d in value}
        return set(value) if isinstance(value, list) else value

    elif orig_tp is dict and isinstance(value, dict):
        args = get_args(tp)
        if len(args) >= 2 and is_dataclass(args[1]):
            return {k: _nested_validate(v, args[1]) for k, v in value.items()}
        elif len(args) >= 2:
            return {k: _validate_single_value(v, args[1]) for k, v in value.items()}
        return value

    elif orig_tp is tuple and isinstance(value, tuple | list):
        args = get_args(tp)
        if args:
            result = []
            for i, item in enumerate(value):
                if i < len(args):
                    if is_dataclass(args[i]):
                        result.append(_nested_validate(item, args[i]))
                    else:
                        result.append(_validate_single_value(item, args[i]))
                else:
                    result.append(item)
            return tuple(result)
        return tuple(value) if isinstance(value, list) else value

    return None


def _validate_single_value(value: Any, tp: Any) -> Any:
    """验证单个值"""
    if origin_is_union(get_origin(tp)):
        success, validated = _validate_union_type(value, tp)
        return validated if success else value
    elif is_dataclass(tp):
        if isinstance(value, dict):
            return _nested_validate(value, tp)
        return value
    elif tp is Path:
        return Path(value)
    elif tp is datetime:
        if isinstance(value, (int, float)):  # noqa: UP038
            return datetime.fromtimestamp(value)
        return datetime.fromisoformat(value)
    elif tp is date:
        if isinstance(value, (int, float)):  # noqa: UP038
            return date.fromtimestamp(value)
        return date.fromisoformat(value)
    elif generic_isinstance(value, tp):
        return value
    else:
        validated = _validate_complex_type(value, tp)
        return validated if validated is not None else value


def _nested_validate(namespace: dict[str, Any], cls_):
    """递归验证嵌套的数据类"""
    result = {}
    types_namespace = merge_cls_and_parent_ns(cls_)

    for field_ in fields(cls_):
        if field_.name not in namespace:
            continue

        tp = _resolve_type(field_.type, types_namespace)
        field_.type = tp
        value = namespace[field_.name]

        if origin_is_union(get_origin(tp)):
            success, validated_value = _validate_union_type(value, tp)
            if success:
                result[field_.name] = validated_value
            continue

        validated_value = _validate_single_value(value, tp)
        result[field_.name] = validated_value

    return cls_(**result)


class BasicConfModelAction(ConfigModelAction[BasicConfModel]):
    # fmt: off
    @classmethod
    def load(cls, data: dict[str, Any], t: type[BasicConfModel]) -> BasicConfModel:
        return _nested_validate(data, t)

    # fmt: on
    @classmethod
    def dump(cls, obj: BasicConfModel) -> dict[str, Any]:
        return asdict(obj)  # type: ignore

    @classmethod
    def keys(cls, obj: BasicConfModel) -> list[str]:
        return [field_.name for field_ in fields(obj)]  # type: ignore

    @classmethod
    def schema(cls, t: type[BasicConfModel], ref_root: str = "/") -> dict[str, Any]:
        return SchemaGenerator.from_dc(t, ref_root)  # type: ignore


__all__ = ["BasicConfModel", "field"]



from dataclasses import dataclass, field
from typing import Generic, TypeVar

from arclet.alconna import Alconna, Arparma
from tarina import Empty

T = TypeVar("T")


@dataclass
class Match(Generic[T]):
    """
    匹配项，表示参数是否存在于 `all_matched_args` 内

    Attributes:
        result (T): 匹配结果
        available (bool): 匹配状态
    """

    result: T
    available: bool


class Query(Generic[T]):
    """
    查询项，表示参数是否可由 `Arparma.query` 查询并获得结果

    Attributes:
        result (T): 查询结果
        available (bool): 查询状态
        path (str): 查询路径
    """

    result: T
    available: bool
    path: str

    def __init__(self, path: str, default: T | type[Empty] = Empty):
        self.path = path
        self.result = default  # type: ignore
        self.available = False

    def __repr__(self):
        return f"Query({self.path}, {self.result})"


@dataclass(frozen=True)
class CommandResult:
    source: Alconna
    result: Arparma
    output: str | None = field(default=None)

    @property
    def matched(self) -> bool:
        return self.result.matched



from dataclasses import field
import random

from arclet.alconna import (
    Alconna,
    Args,
    Arparma,
    CommandMeta,
    Field,
    Option,
    Subcommand,
    SubcommandResult,
    command_manager,
    store_true,
)
from tarina import lang

from arclet.entari import BasicConfModel, Session, command, metadata, plugin_config


class Config(BasicConfModel):
    help_command: str = "help"
    """帮助指令的指令名"""
    help_alias: list[str] = field(default_factory=lambda: ["帮助", "命令帮助"])
    """帮助指令的别名"""
    help_all_alias: list[str] = field(default_factory=lambda: ["所有帮助", "所有命令帮助"])
    """帮助指令的'所有帮助'别名"""
    page_size: int | None = None
    """帮助结果每页显示的命令数量"""


config = plugin_config(Config)


metadata(
    "帮助",
    author=[{"name": "RF-Tar-Railt", "email": "rf_tar_railt@qq.com"}],
    description="展示所有命令的帮助(菜单)",
    readme="""
# 帮助指令

该插件提供查看所有命令帮助的功能等。

## 配置

- `help_command`: 帮助指令的指令名，默认为 `help`
- `help_alias`: 帮助指令的别名，默认为 `["帮助", "命令帮助"]`
- `help_all_alias`: 帮助指令的'所有帮助'别名，默认为 `["所有帮助", "所有命令帮助"]`
- `page_size`: 帮助结果每页显示的命令数量，默认为 `None`，表示不分页

## 使用

- 查看所有命令帮助: `help` 或 `帮助`
- 查看指定命令帮助: `help [id|name]` 或 `帮助 [id|name]`
- 查看指定页数的命令帮助: `help --page [index]` 或 `帮助 --page [index]`
- 查看命令所属命名空间: `help [id|name] --namespace` 或 `帮助 [id|name] --namespace`
- 列出所有命令所属命名空间: `help --namespace --list` 或 `帮助 --namespace --list`
- 查看隐藏命令: `help --hide` 或 `帮助 --hide`
""",
    config=Config,
)


help_cmd = Alconna(
    config.help_command,
    Args[
        "query#选择某条命令的id或者名称查看具体帮助;/?",
        str,
        Field(
            "-1",
            completion=lambda: f"试试 {random.randint(0, len(command_manager.get_commands()))}",
            unmatch_tips=lambda x: f"预期输入为某个命令的id或者名称，而不是 {x}\n例如：/帮助 0",
        ),
    ],
    Option(
        "--page",
        Args["index", int],
        help_text="查看指定页数的命令帮助",
    ),
    Subcommand(
        "--namespace",
        Args["target?;#指定的命名空间", str],
        Option("--list", help_text="列出所有命名空间", action=store_true, default=False),
        alias=["-N", "命名空间"],
        help_text="是否列出命令所属命名空间",
    ),
    Option("--hide", alias=["-H", "隐藏"], help_text="是否列出隐藏命令", action=store_true, default=False),
    meta=CommandMeta(
        description="显示所有命令帮助",
        usage="可以使用 --hide 参数来显示隐藏命令，使用 -P 参数来显示命令所属插件名称",
        example=f"${config.help_command} 1",
    ),
)

for alias in set(config.help_alias):
    help_cmd.shortcut(alias, {"prefix": True, "fuzzy": False})
for alias in set(config.help_all_alias):
    help_cmd.shortcut(alias, {"args": ["--hide"], "prefix": True, "fuzzy": False})


def help_cmd_handle(arp: Arparma, interactive: bool = False):
    is_namespace = arp.query[SubcommandResult]("namespace")
    page = arp.query[int]("page.index", 1)
    target_namespace = is_namespace.args.get("target") if is_namespace else None
    cmds = [
        i
        for i in command_manager.get_commands(target_namespace or "")
        if not i.meta.hide or arp.query[bool]("hide.value", False)
    ]
    if is_namespace and is_namespace.options["list"].value and not target_namespace:
        namespaces = {i.namespace: 0 for i in cmds}
        return "\n".join(
            f" 【{str(index).rjust(len(str(len(namespaces))), '0')}】{n}" for index, n in enumerate(namespaces.keys())
        )

    help_names = set()
    for i in cmds:
        help_names.update(i.namespace_config.builtin_option_name["help"])

    footer = lang.require("manager", "help_footer").format(help="|".join(sorted(help_names, key=lambda x: len(x))))
    show_namespace = is_namespace and not is_namespace.options["list"].value and not target_namespace
    if (query := arp.all_matched_args["query"]) != "-1":
        if query.isdigit():
            index = int(query)
            if index < 0 or index >= len(cmds):
                return "查询失败！"
            slot = cmds[index]
        elif not (slot := next((i for i in cmds if query == i.command), None)):
            command_string = "\n".join(
                (
                    f"【{str(index).rjust(len(str(len(cmds))), '0')}】"
                    f"{f'{slot.namespace}::' if show_namespace else ''}{slot.header_display} : "
                    f"{slot.meta.description}"
                )
                for index, slot in enumerate(cmds)
                if query in str(slot.command)
            )
            if not command_string:
                return "查询失败！"
            return f"{command_string}\n{footer}"
        return slot.get_help()

    if not config.page_size:
        header = lang.require("manager", "help_header")
        command_string = "\n".join(
            (
                f" 【{str(index).rjust(len(str(len(cmds))), '0')}】"
                f"{f'{slot.namespace}::' if show_namespace else ''}{slot.header_display} : "
                f"{slot.meta.description}"
            )
            for index, slot in enumerate(cmds)
        )
        return f"{header}\n{command_string}\n{footer}"

    max_page = len(cmds) // config.page_size + 1
    if page < 1 or page > max_page:
        page = 1
    max_length = config.page_size
    if interactive:
        footer += "\n" + "输入 '<', 'a' 或 '>', 'd' 来翻页"

    def _(_page: int):
        header = (
            lang.require("manager", "help_header")
            + "\t"
            + lang.require("manager", "help_pages").format(current=_page, total=max_page)
        )
        command_string = "\n".join(
            (
                f" 【{str(index).rjust(len(str(_page * max_length)), '0')}】"
                f"{f'{slot.namespace}::' if show_namespace else ''}{slot.header_display} : "
                f"{slot.meta.description}"
            )
            for index, slot in enumerate(
                cmds[(_page - 1) * max_length : _page * max_length], start=(_page - 1) * max_length
            )
        )
        return f"{header}\n{command_string}\n{footer}"

    if not interactive:
        return _(page)

    def generator(_page: int):
        while True:
            resp = yield _(_page)
            if resp == "a" or resp == "<":
                _page -= 1
                if _page < 1:
                    _page = max_page
            elif resp == "d" or resp == ">":
                _page += 1
                if _page > max_page:
                    _page = 1
            else:
                return

    return generator(page)


disp = command.mount(help_cmd)


@disp.on_execute()
async def help_exec(arp: Arparma):
    return help_cmd_handle(arp)


@disp.handle
async def help_handle(arp: Arparma, session: Session):
    resp = help_cmd_handle(arp, True)
    if isinstance(resp, str):
        return await session.send(resp)

    async def receiver(sess1: Session):
        if (ans := str(sess1.content).strip().lower()) in {"<", "a", ">", "d"}:
            return ans

    msg = await session.prompt(receiver, next(resp), timeout=15)
    while msg:
        try:
            msg = await session.prompt(resp.send(msg), timeout=15)
        except StopIteration:
            return



# This file is @generated by tarina.lang CLI tool
# It is not intended for manual editing.

from pathlib import Path

from tarina.lang import lang

lang.load(Path(__file__).parent)

from .model import Lang as Lang



import ast
from collections.abc import Mapping
from dataclasses import Field
import inspect
from types import MappingProxyType
from typing import cast

_DESC_CACHE = {}


class GetattrDict:
    def __init__(self, source: Mapping):
        self._source = source

    def __getitem__(self, item):
        return self._source[item]

    def __getattr__(self, item):
        try:
            return self._source[item]
        except KeyError as e:
            raise AttributeError(f"{item} not found") from e


def cleanup_src(src: str) -> str:
    lines = src.expandtabs().split("\n")
    margin = len(lines[0]) - len(lines[0].lstrip())
    for i in range(len(lines)):
        lines[i] = lines[i][margin:]
    return "\n".join(lines)


def store_field_description(cls: type, fields: dict[str, Field]) -> None:
    if cls in _DESC_CACHE:
        return
    try:
        node: ast.ClassDef = cast(ast.ClassDef, ast.parse(cleanup_src(inspect.getsource(cls))).body[0])
    except (TypeError, OSError):  # NOTE: for REPL.
        return
    for i, stmt in enumerate(node.body):
        name: str | None = None
        if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
            name = stmt.target.id
        if (
            name in fields
            and i + 1 < len(node.body)
            and isinstance((doc_expr := node.body[i + 1]), ast.Expr)
            and isinstance((doc_const := doc_expr.value), ast.Constant)
            and isinstance(doc_string := doc_const.value, str)
            and "description" not in (field := fields[name]).metadata
        ):
            field.metadata = MappingProxyType({**field.metadata.copy(), "description": inspect.cleandoc(doc_string)})
    _DESC_CACHE[cls] = True



import asyncio
from collections.abc import Callable
import inspect
import itertools
from os import PathLike
from pathlib import Path
import sys
from typing import TYPE_CHECKING, Any, Literal, TypeVar, overload

from arclet.letoderea import Subscriber, on, publish
from arclet.letoderea.typing import Resultable
from tarina import init_spec
from tarina.tools import nest_obj_update

from ..config import EntariConfig, config_model_keys, config_model_validate
from ..event.config import ConfigReload
from ..event.lifespan import Ready
from ..event.plugin import PluginLoadedFailed
from ..exceptions import RegisterNotInPluginError, ReusablePluginError, StaticPluginDispatchError
from ..logger import escape_tag, log
from .model import PluginMetadata as PluginMetadata
from .model import RootlessPlugin as RootlessPlugin
from .model import TS, Plugin, PluginDispatcher
from .model import current_plugin
from .model import keeping as keeping
from .module import import_plugin
from .module import package as package
from .module import requires as requires
from .service import plugin_service

T = TypeVar("T")


@overload
def get_plugin(depth: int = 0) -> Plugin: ...


@overload
def get_plugin(depth: int = 0, *, optional: Literal[True]) -> Plugin | None: ...


def get_plugin(depth: int = 0, *, optional: bool = False) -> Plugin | None:
    """获取当前插件上下文

    Args:
        depth (int, optional): 获取的深度，默认为0，表示当前插件上下文. Defaults to 0.
        optional (bool, optional): 是否允许返回None，默认为False. Defaults to False.
    Raises:
        ValueError: 如果深度超出范围
        LookupError: 如果没有找到插件上下文
    Returns:
        Plugin | None: 当前插件上下文，如果没有找到且optional为True则返回None
    """
    if plugin := current_plugin.get(None):
        return plugin
    current_frame = inspect.currentframe()
    if current_frame is None:
        if optional:
            return None
        raise ValueError("Depth out of range")
    frame = current_frame
    d = depth + 1
    while d > 0:
        frame = frame.f_back
        if frame is None:
            if optional:
                return None
            raise ValueError("Depth out of range")
        d -= 1
    locals_ = frame.f_locals
    if "__plugin__" in locals_:
        return locals_["__plugin__"]
    globals_ = frame.f_globals
    if "__plugin__" in globals_:
        return globals_["__plugin__"]
    if optional:
        return None
    raise LookupError("no plugin context found")


def get_plugins(subplugged: bool = False):
    """获取所有插件

    Args:
        subplugged (bool, optional): 是否包含子插件. Defaults to False.
    """
    if subplugged:
        return list(plugin_service.plugins.values())
    return [p for pid, p in plugin_service.plugins.items() if pid not in plugin_service._subplugined]


def get_plugin_subscribers(plug: Plugin | str | None = None) -> list[Subscriber]:
    """获取指定插件的所有订阅者"""
    if isinstance(plug, Plugin):
        plg = plug
    elif plug in plugin_service.plugins:
        plg = plugin_service.plugins[plug]
    else:
        plg = get_plugin(1)
    return [s[0] for s in plg._scope.subscribers.values()]


def get_all_subscribers():
    """获取所有插件的所有订阅者"""
    return list(itertools.chain.from_iterable(get_plugin_subscribers(plug) for plug in plugin_service.plugins.values()))


def get_plugin_references(plug: Plugin | str | None = None) -> set[str]:
    """获取指定插件引用的所有插件

    Args:
        plug (Plugin | str | None, optional): 插件对象或插件名称. Defaults to None.
    """
    if isinstance(plug, Plugin):
        plg = plug
    elif plug in plugin_service.plugins:
        plg = plugin_service.plugins[plug]
    else:
        plg = get_plugin(1)
    return plugin_service.references.get(plg.id, set()).copy()


def get_plugin_referents(plug: Plugin | str | None = None) -> set[str]:
    """获取引用了指定插件的所有插件

    Args:
        plug (Plugin | str | None, optional): 插件对象或插件名称. Defaults to None.
    """
    if isinstance(plug, Plugin):
        plg = plug
    elif plug in plugin_service.plugins:
        plg = plugin_service.plugins[plug]
    else:
        plg = get_plugin(1)
    return plugin_service.referents.get(plg.id, set()).copy()


@overload
def dispatch(event: type[Resultable[T]], name: str | None = None) -> PluginDispatcher[T]: ...
@overload
def dispatch(event: type[Any], name: str | None = None) -> PluginDispatcher[Any]: ...


def dispatch(event: type, name: str | None = None) -> PluginDispatcher:
    """对当前插件创建一个事件分发"""
    return get_plugin(1).dispatch(event, name=name)


def load_plugin(
    path: str, config: dict | None = None, recursive_guard: set[str] | None = None, prelude: bool = False
) -> Plugin | None:
    """
    以导入路径方式加载模块

    Args:
        path (str): 模块路径
        config (dict): 模块配置
        recursive_guard (set[str]): 递归保护
        prelude (bool): 是否为前置插件
    """
    if config is None:
        config = EntariConfig.instance.plugin.get(path, {})
    config["$path"] = path
    if prelude:
        config["$static"] = True
    if recursive_guard is None:
        recursive_guard = set()
    path = path.replace("::", "arclet.entari.builtins.")
    while path in plugin_service._subplugined:
        path = plugin_service._subplugined[path]
    if path in plugin_service._apply:
        log.plugin.trace(f"loaded rootless plugin <y>{path!r}</y>")
        return plugin_service._apply[path][0](config)
    if plug := find_plugin(path):
        plugin_service._direct_plugins.add(plug.path)
        return plug
    try:
        if pref := config.get("$prefix"):
            path = f"{pref if isinstance(pref, str) else 'entari_plugin_'}{path}"
        mod = import_plugin(path, config=config)
        if not pref and not mod and not path.count("."):
            path1 = f"entari_plugin_{path}"
            mod = import_plugin(path1, config=config)
        if not mod:
            log.plugin.error(f"cannot found plugin <blue>{path!r}</blue>")
            publish(PluginLoadedFailed(path))
            return
        plugin_service._direct_plugins.add(mod.__name__)
        if mod.__name__ in plugin_service.referents and plugin_service.referents[mod.__name__]:
            referents = plugin_service.referents[mod.__name__].copy()
            # plugin_service.referents[mod.__name__].clear()
            for referent in referents:
                if referent in recursive_guard:
                    continue
                if referent.startswith(mod.__name__):
                    continue
                if referent in plugin_service.plugins:
                    plugin_service.referents[mod.__name__].discard(referent)
                    log.plugin.debug(f"reloading <y>{escape_tag(mod.__name__)}</y>'s referent <y>{referent!r}</y>")
                    unload_plugin(referent)
                if not (plug := load_plugin(referent)):
                    plugin_service.referents[mod.__name__].add(referent)
                else:
                    publish(Ready(), plug._scope)
                    recursive_guard.add(referent)

        return mod.__plugin__
    except (ImportError, RegisterNotInPluginError, ReusablePluginError, StaticPluginDispatchError):
        return
    except Exception as e:
        log.plugin.exception(f"failed to load plugin <blue>{path!r}</blue>: {e}", exc_info=e)
        publish(PluginLoadedFailed(path))
        return


def load_plugins(dir_: str | PathLike | Path):
    """加载指定目录下的所有插件"""
    path = dir_ if isinstance(dir_, Path) else Path(dir_)
    if not path.is_dir():
        raise NotADirectoryError(f"{path} is not a directory")
    path: Path = path.resolve()  # .relative_to(Path.cwd())
    syspaths = [Path(p).resolve() for p in sys.path if p]
    prefixes = [p for p in syspaths if path.is_relative_to(p)]
    if prefixes:
        prefix = max(prefixes, key=lambda p: len(p.parts))
    else:
        prefix = Path.cwd()
    for p in path.iterdir():
        if p.suffix in (".py", "") and p.stem not in {"__init__", "__pycache__"}:
            p = p.resolve().relative_to(prefix)
            plg = ".".join(p.parts[:-1:1]) + "." + p.stem
            load_plugin(plg)


if TYPE_CHECKING:

    @init_spec(PluginMetadata)
    def metadata(data: PluginMetadata) -> None:
        """声明当前插件的元数据"""

else:

    def metadata(*args, **kwargs):
        """声明当前插件的元数据"""
        get_plugin(1).metadata = PluginMetadata(*args, **kwargs)  # type: ignore


_C = TypeVar("_C")


@overload
def plugin_config() -> dict[str, Any]: ...


@overload
def plugin_config(model_type: type[_C], bind: bool = False) -> _C: ...


def plugin_config(model_type: type[_C] | None = None, bind: bool = False):
    """获取当前插件的配置

    Args:
        model_type (type[_C], optional): 配置模型类型. Defaults to None.
        bind (bool, optional): 是否将配置模型与配置绑定，绑定后配置模型的修改会更新配置文件,
            而配置文件的修改则直接作用在配置模型上，不再重载整个插件. Defaults to False.
    """
    plugin = get_plugin(1)
    if model_type:
        obj = config_model_validate(model_type, plugin.config)
        if bind:
            plugin_key = plugin._config_key

            def _reload(event: ConfigReload):
                if event.scope != "plugin":
                    return
                if event.key != plugin_key:
                    return
                new = config_model_validate(model_type, event.value)
                nest_obj_update(obj, new, config_model_keys(new))
                return True

            sub = plugin._scope.register(_reload, event=ConfigReload)
            proxy = EntariConfig.instance.bind(plugin_key, obj)
            plugin.collect(lambda: delattr(proxy, "_Proxy__origin"), sub.dispose)
            return proxy
        return obj
    return plugin.config


get_config = plugin_config


def declare_static():
    """声明当前插件为静态插件"""
    plugin = get_plugin(1)
    plugin.is_static = True
    if plugin._scope.subscribers:
        raise StaticPluginDispatchError("static plugin cannot dispatch events")


def add_service(serv: TS | type[TS]) -> TS:
    """添加一个服务到当前插件"""
    return get_plugin(1).service(serv)


def collect_disposes(*disposes: Callable[[], None]):
    """收集副作用回收函数"""
    return get_plugin(1).collect(*disposes)


def restore():
    """回收该插件收集的所有副作用"""
    return get_plugin(1).restore()


def find_plugin(name: str) -> Plugin | None:
    """根据插件名称查找插件"""
    if name in plugin_service.plugins:
        return plugin_service.plugins[name]
    if not name.count(".") and f"entari_plugin_{name}" in plugin_service.plugins:
        return plugin_service.plugins[f"entari_plugin_{name}"]


def find_plugin_by_file(file: str) -> Plugin | None:
    path = Path(file).resolve()
    for plugin in plugin_service.plugins.values():
        if plugin.module.__file__ == str(path):
            return plugin
        if plugin.module.__file__ and Path(plugin.module.__file__).parent == path:
            return plugin
        path1 = Path(path)
        while path1.parent != path1:
            if str(path1) == plugin.module.__file__:
                return plugin
            path1 = path1.parent
    return None


def unload_plugin(plugin: str):
    plugin = plugin.replace("::", "arclet.entari.builtins.")
    while plugin in plugin_service._subplugined:
        plugin = plugin_service._subplugined[plugin]
    if not (_plugin := find_plugin(plugin)):
        return False
    _plugin.dispose()
    return True


async def unload_plugin_async(plugin: str):
    plugin = plugin.replace("::", "arclet.entari.builtins.")
    while plugin in plugin_service._subplugined:
        plugin = plugin_service._subplugined[plugin]
    if not (_plugin := find_plugin(plugin)):
        return False
    if tasks := _plugin.dispose():
        await asyncio.wait(tasks)
    return True


async def enable_plugin(plugin: str):
    """启用指定插件"""
    plugin = plugin.replace("::", "arclet.entari.builtins.")
    while plugin in plugin_service._subplugined:
        plugin = plugin_service._subplugined[plugin]
    if not (_plugin := find_plugin(plugin)):
        return False
    if tasks := _plugin.enable():
        await asyncio.wait(tasks)
    return True


async def disable_plugin(plugin: str):
    """禁用指定插件"""
    plugin = plugin.replace("::", "arclet.entari.builtins.")
    while plugin in plugin_service._subplugined:
        plugin = plugin_service._subplugined[plugin]
    if not (_plugin := find_plugin(plugin)):
        return False
    if tasks := _plugin.disable():
        await asyncio.wait(tasks)
    return True


listen = on



from arclet.alconna import Argv, argv_config, set_default_argv_type, set_namespace_argv_type
from satori import Text

from ..message import MessageChain


class MessageArgv(Argv[MessageChain]):
    @staticmethod
    def generate_token(data: list) -> int:
        return hash("".join(i.__repr__() for i in data))


set_default_argv_type(MessageArgv)
set_namespace_argv_type("Entari", MessageArgv)


argv_config(
    MessageArgv,
    filter_out=[],
    to_text=lambda x: x.text if x.__class__ is Text else None,
    converter=MessageChain,
)



import asyncio
from asyncio.events import _get_running_loop  # type: ignore
from collections.abc import Callable
from datetime import datetime, timedelta
from traceback import print_exc
from typing import Literal

from arclet.letoderea import Scope, Subscriber, make_event
from arclet.letoderea.typing import Contexts
from launart import Launart, Service, any_completed
from launart.status import Phase

from .plugin import RootlessPlugin, get_plugin, metadata


@make_event(name="entari.event/internal/schedule")
class _ScheduleEvent:
    pass


scope = Scope.of("entari.scheduler")
contexts: Contexts = {"$event": _ScheduleEvent()}  # type: ignore


class TimerTask:
    def __init__(self, supplier: Callable[[], timedelta], sub: Subscriber):
        self.sub = sub
        self.supplier = supplier
        self.handle = None
        self.available = True

    def start(self, queue: asyncio.Queue):
        loop = asyncio.get_running_loop()
        self.handle = loop.call_later(self.supplier().total_seconds(), queue.put_nowait, self)

    def cancel(self):
        if self.handle and not self.handle.cancelled():
            self.handle.cancel()


class Scheduler(Service):

    def __init__(self):
        super().__init__()
        self.queue: asyncio.Queue[TimerTask] = asyncio.Queue()
        self.tasks: dict[str, TimerTask] = {}

    @property
    def required(self) -> set[str]:
        return set()

    @property
    def stages(self) -> set[Phase]:
        return {"preparing", "blocking", "cleanup"}

    async def fetch(self):
        while True:
            task = await self.queue.get()
            if not task.available:
                del self.tasks[task.sub.id]
                continue
            task.start(self.queue)
            if task.sub.available:
                try:
                    await task.sub.handle(contexts.copy())
                except Exception:
                    print_exc()

    def schedule(self, time_fn: Callable[[], timedelta], once: bool = False):
        """
        设置一个定时任务

        Args:
            time_fn: 用于提供定时间隔的函数
            once: 是否只执行一次
        """

        def wrapper(func: Callable):
            plg = get_plugin(1, optional=True)
            if plg:
                sub = plg.dispatch(_ScheduleEvent).handle(func, once=once)
            else:
                sub = scope.register(func, _ScheduleEvent, once=once)
            task = self.tasks[sub.id] = TimerTask(time_fn, sub)

            def _dispose(_):
                task.available = False

            sub._attach_disposes(_dispose)

            if _get_running_loop():
                self.tasks[sub.id].start(self.queue)
            return sub

        return wrapper

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            for task in self.tasks.values():
                task.start(self.queue)

        async with self.stage("blocking"):
            sigexit_task = asyncio.create_task(manager.status.wait_for_sigexit())
            fetch_task = asyncio.create_task(self.fetch())
            done, pending = await any_completed(sigexit_task, fetch_task)
            if sigexit_task in done:
                fetch_task.cancel()

        async with self.stage("cleanup"):
            for task in self.tasks.values():
                task.cancel()

    id = "entari.scheduler"


scheduler = service = Scheduler()
schedule = scheduler.schedule


@RootlessPlugin.apply("scheduler", default=True)
def _(plg: RootlessPlugin):
    metadata("Simple Scheduler with interval / crontab task", ["RF-Tar-Railt <rf_tar_railt@qq.com>"])

    plg.service(service)


class timer:
    @staticmethod
    def every_second():
        """每秒执行一次"""
        return lambda: timedelta(seconds=1)

    @staticmethod
    def every_seconds(seconds: int):
        """每 seconds 秒执行一次

        Args:
            seconds (int): 距离下一次执行的时间间隔, 单位为秒
        """
        if seconds > 59 or seconds < 1:
            raise ValueError
        return lambda: timedelta(seconds=seconds)

    @staticmethod
    def every_minute():
        """每分钟执行一次"""
        return lambda: timedelta(minutes=1)

    @staticmethod
    def every_minutes(minutes: int):
        """每 minutes 分钟执行一次

        Args:
            minutes (int): 距离下一次执行的时间间隔, 单位为分
        """
        if minutes > 59 or minutes < 1:
            raise ValueError
        return lambda: timedelta(minutes=minutes)

    @staticmethod
    def every_hour():
        """每小时执行一次"""
        return lambda: timedelta(hours=1)

    @staticmethod
    def every_hours(hours: int):
        """每 hours 小时执行一次

        Args:
            hours (int): 距离下一次执行的时间间隔, 单位为小时
        """
        if hours > 23 or hours < 1:
            raise ValueError
        return lambda: timedelta(hours=hours)

    @staticmethod
    def every_week():
        """每隔一周执行一次"""
        return lambda: timedelta(weeks=1)

    @staticmethod
    def every_weeks(weeks: int):
        """每 weeks 周执行一次

        Args:
            weeks (int): 距离下一次执行的时间间隔, 单位为周
        """
        if weeks > 52 or weeks < 1:
            raise ValueError
        return lambda: timedelta(weeks=weeks)

    @staticmethod
    def every_day():
        """每隔一天执行一次"""
        return lambda: timedelta(days=1)

    @staticmethod
    def every_days(days: int):
        """每 days 天执行一次

        Args:
            days (int): 距离下一次执行的时间间隔, 单位为天
        """
        if days > 31 or days < 1:
            raise ValueError
        return lambda: timedelta(days=days)

    @staticmethod
    def crontab(cron_str: str):
        """根据 cron 表达式执行

        Args:
            cron_str (str): cron 表达式
        """
        try:
            from croniter import croniter
        except ModuleNotFoundError:
            raise ImportError(
                "Please install `croniter` first. Install with `pip install arclet-entari[cron]`"
            ) from None

        it = croniter(cron_str, datetime.now(), datetime)

        return lambda iter=it: iter.get_next(datetime) - datetime.now()


def cron(pattern: str):
    """使用 cron 表达式设置一个定时任务"""
    return service.schedule(timer.crontab(pattern))


def every(value: int = 1, mode: Literal["second", "minute", "hour"] = "second"):
    """依据 mode 设置一个定时任务"""
    _TIMER_MAPPING = {
        "second": timer.every_seconds,
        "minute": timer.every_minutes,
        "hour": timer.every_hours,
    }
    return service.schedule(_TIMER_MAPPING[mode](value))


def invoke(delay: float):
    """延迟 delay 秒执行"""
    return service.schedule(lambda: timedelta(seconds=delay), once=True)


__all__ = ["scheduler", "schedule", "timer", "cron", "every", "invoke"]



import ast
from collections.abc import Sequence
from importlib import _bootstrap, _bootstrap_external  # type: ignore
from importlib.abc import MetaPathFinder
from importlib.machinery import ExtensionFileLoader, PathFinder, SourceFileLoader
from importlib.util import module_from_spec, resolve_name
from io import BytesIO
import re
import sys
import tokenize
from types import ModuleType

from arclet.letoderea import publish
from arclet.letoderea.scope import scope_ctx

from ..config import EntariConfig
from ..event.lifespan import Ready
from ..event.plugin import PluginLoadedFailed, PluginLoadedSuccess
from ..exceptions import RegisterNotInPluginError, ReusablePluginError, StaticPluginDispatchError
from ..logger import log
from .model import Plugin, PluginMetadata, current_plugin
from .service import plugin_service

_SUBMODULE_WAITLIST: dict[str, set[str]] = {}
_ENSURE_IS_PLUGIN: set[str] = set()
_IMPORTING = set()

PLUGIN_PAT = re.compile(r"entari:\s*plugin")
SUBPLUGIN_PAT = re.compile(r"entari:\s*(?:package|subplugin)")
NAMESPACE_PAT = re.compile(r"entari:\s*namespace")


def package(*names: str):
    """手动指定特定模块作为插件的子模块"""
    if not (plugin := current_plugin.get(None)):
        raise LookupError("no plugin context found")
    _SUBMODULE_WAITLIST.setdefault(plugin.module.__name__, set()).update(names)


def requires(*names: str):
    """手动指定哪些模块是插件"""
    _ENSURE_IS_PLUGIN.update(name.replace("::", "arclet.entari.builtins.") for name in names)


def _ensure_plugin(names: list[str], sub: bool, pid: str, pname: str, prefix=""):
    for name in names:
        if sub:
            _SUBMODULE_WAITLIST.setdefault(pname, set()).add(f"{prefix}{name}")
        else:
            _ENSURE_IS_PLUGIN.add(f"{prefix}{name}")
            plugin_service.referents.setdefault(f"{prefix}{name}", set()).add(pid)
            plugin_service.references.setdefault(pname, set()).add(f"{prefix}{name}")
        _IMPORTING.add(f"{prefix}{name}")


# fmt: off
class _Visitor(ast.NodeVisitor):
    def __init__(self, pid: str, pname: str, path: str, plg_lineno: list[int], sub_lineno: list[int], ns_lineno: list[int]):  # noqa: E501
        self.pid = pid
        self.pname = pname
        self.path = path
        self.signed_plugin_lineno = plg_lineno
        self.signed_subplugin_lineno = sub_lineno
        self.signed_namespace_lineno = ns_lineno
        self.signed_namespace_modules = set()
        self.type_checking_stack = []
        self.typing_aliases = {"typing", "typing_extensions"}  # 跟踪typing模块的别名

    def visit_Import(self, node: ast.Import):
        # 跟踪typing模块的导入别名
        for alias in node.names:
            if alias.name in ("typing", "typing_extensions"):
                self.typing_aliases.add(alias.asname or alias.name)

        if self._in_type_checking():
            return
        if node.lineno in self.signed_plugin_lineno or all(x.name in _ENSURE_IS_PLUGIN for x in node.names):
            _ensure_plugin([alias.name for alias in node.names], False, self.pid, self.pname)
        elif node.lineno in self.signed_subplugin_lineno or all(x.name in _SUBMODULE_WAITLIST.get(self.pname, ()) for x in node.names):  # noqa: E501
            _ensure_plugin([alias.name for alias in node.names], True, self.pid, self.pname)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        name = self.pname
        if self._in_type_checking():
            return
        if node.module is None:  # from . import xxx
            _ensure_plugin([alias.name for alias in node.names], node.lineno not in self.signed_plugin_lineno, self.pid, name, f"{name}.")  # noqa: E501
        elif node.level == 0:  # from xxx import xxx
            if node.module in (*sys.builtin_module_names, *getattr(sys, "stdlib_module_names", [])):
                return
            if node.lineno in self.signed_plugin_lineno or node.module in _ENSURE_IS_PLUGIN:
                _ensure_plugin([node.module], False, self.pid, name)
            elif node.lineno in self.signed_subplugin_lineno or node.module in _SUBMODULE_WAITLIST.get(name, ()):
                _ensure_plugin([node.module], True, self.pid, name)
        elif node.level == 1:  # from .xxx import xxx
            if node.lineno in self.signed_plugin_lineno:
                pname = f"{name}.{node.module}"
                _ensure_plugin([alias.name for alias in node.names], False, self.pid, pname, f"{pname}.")  # noqa: E501
            elif node.lineno in self.signed_namespace_lineno or node.module in self.signed_namespace_modules:
                self.signed_namespace_modules.add(node.module)
                pname = f"{name}.{node.module}"
                _ensure_plugin([alias.name for alias in node.names], False, self.pid, pname, f"{pname}.")
            prefix = name if self.path.endswith("__init__.py") else name.rpartition(".")[0]
            _ensure_plugin([node.module], True, self.pid, prefix, f"{prefix}.")
        else:  # from ..xxx import xxx
            prefix = ".".join(name.split(".")[: -node.level + 1 if self.path.endswith("__init__.py") else -node.level])
            if not prefix:  # relative import beyond top-level package
                return
            if prefix not in plugin_service.plugins and prefix not in _ENSURE_IS_PLUGIN:
                is_sub = prefix in _SUBMODULE_WAITLIST.get(name, ())
                pname = name
            else:
                is_sub = True
                pname = prefix
            _ensure_plugin([node.module], is_sub, self.pid, pname, f"{prefix}.")

    def visit_If(self, node: ast.If):
        is_type_checking = self._is_type_checking(node)
        if is_type_checking:
            self.type_checking_stack.append(True)
        self.generic_visit(node)
        if is_type_checking:
            self.type_checking_stack.pop()

    def visit_Call(self, node):
        # 寻找 `requires(...)` 和 `package(...)` 调用
        if isinstance(node.func, ast.Name) and node.func.id == "requires":
            names = []
            for arg in node.args:
                if isinstance(arg, ast.Constant) and isinstance(arg.value, str):
                    names.append(arg.value)
            if names:
                _ensure_plugin(names, False, self.pid, self.pname)
        elif isinstance(node.func, ast.Name) and node.func.id == "package":
            names = []
            for arg in node.args:
                if isinstance(arg, ast.Constant) and isinstance(arg.value, str):
                    names.append(arg.value)
            if names:
                _ensure_plugin(names, True, self.pid, self.pname)
        self.generic_visit(node)

    def _in_type_checking(self) -> bool:
        return len(self.type_checking_stack) > 0

    def _is_type_checking(self, node: ast.If) -> bool:
        test = node.test
        if isinstance(test, ast.Name) and test.id == "TYPE_CHECKING":
            return True
        if isinstance(test, ast.Attribute):
            if isinstance(test.value, ast.Name) and test.value.id in self.typing_aliases and test.attr == "TYPE_CHECKING":  # noqa: E501
                return True
        if isinstance(test, ast.Compare):
            left = test.left
            if isinstance(left, ast.Name) and left.id == "TYPE_CHECKING":
                return True
            elif isinstance(left, ast.Attribute):
                if isinstance(left.value, ast.Name) and left.value.id in self.typing_aliases and left.attr == "TYPE_CHECKING":  # noqa: E501
                    return True
        return False
# fmt: on


class PluginLoader(SourceFileLoader):
    def __init__(self, fullname: str, path: str, plugin_id: str, parent_plugin_id: str | None = None) -> None:
        self.loaded = False
        self.plugin_id = plugin_id
        self.parent_plugin_id = parent_plugin_id
        super().__init__(fullname, path)

    def get_code(self, fullname):
        """Concrete implementation of InspectLoader.get_code.

        Reading of bytecode requires path_stats to be implemented. To write
        bytecode, set_data must also be implemented.

        """
        source_path = self.get_filename(fullname)
        source_bytes = None
        if source_bytes is None:
            source_bytes = self.get_data(source_path)
        code_object = self.source_to_code(source_bytes, source_path)
        return code_object

    def source_to_code(self, data, path="<string>"):
        """Return the code object compiled from source.

        The 'data' argument can be any object type that compile() supports.
        """
        if not isinstance(data, bytes):
            return _bootstrap._call_with_frames_removed(  # type: ignore
                compile, data, path, "exec", dont_inherit=True, optimize=-1
            )
        name = self.name
        plg_lineno = []
        sub_lineno = []
        ns_lineno = []
        for token in tokenize.tokenize(BytesIO(data).readline):
            if token.type == tokenize.COMMENT:
                if PLUGIN_PAT.search(token.string):
                    plg_lineno.append(token.start[0])
                elif SUBPLUGIN_PAT.search(token.string):
                    sub_lineno.append(token.start[0])
                elif NAMESPACE_PAT.search(token.string):
                    ns_lineno.append(token.start[0])

        try:
            nodes = ast.parse(data, type_comments=True)
        except SyntaxError:
            return _bootstrap._call_with_frames_removed(  # type: ignore
                compile, data, path, "exec", dont_inherit=True, optimize=-1
            )
        visitor = _Visitor(self.plugin_id, name, path, plg_lineno, sub_lineno, ns_lineno)
        visitor.visit(nodes)

        return _bootstrap._call_with_frames_removed(  # type: ignore
            compile, nodes, path, "exec", dont_inherit=True, optimize=-1
        )

    def create_module(self, spec) -> ModuleType | None:
        if self.name in plugin_service.plugins:
            self.loaded = True
            return plugin_service.plugins[self.name].proxy()
        if self.name in plugin_service._subplugined:
            self.loaded = True
            return plugin_service.plugins[plugin_service._subplugined[self.name]].subproxy(self.name)
        if (
            any((k.startswith(self.name) and k.rfind("@") != -1) for k in plugin_service.plugins)
            and self.plugin_id.rfind("@") == -1
        ):
            raise ReusablePluginError(f"reusable plugin {self.name!r} cannot be imported directly")
        return super().create_module(spec)

    def exec_module(self, module: ModuleType, config: dict[str, str] | None = None) -> None:
        is_sub = False
        if plugin := plugin_service.plugins.get(self.parent_plugin_id) if self.parent_plugin_id else None:
            plugin.subplugins.add(module.__name__)
            plugin_service._subplugined[module.__name__] = plugin.id
            is_sub = True

        if self.loaded:
            return

        if config is None:
            key = module.__name__
            if key.startswith("arclet.entari.builtins.") and f"::{key[23:]}" in EntariConfig.instance.plugin:
                key = f"::{key[23:]}"
            elif key.startswith("entari_plugin_") and f"{key[14:]}" in EntariConfig.instance.plugin:
                key = f"{key[14:]}"
            config = EntariConfig.instance.plugin.get(key, {})
            config["$path"] = key
            if key in EntariConfig.instance.prelude_plugin:
                config["$static"] = True  # type: ignore
        # create plugin before executing
        plugin = Plugin(self.plugin_id, module, config=config)
        # for `dataclasses` module
        sys.modules[module.__name__] = plugin.proxy()  # type: ignore
        setattr(module, "__plugin__", plugin)

        # enter plugin context
        token = current_plugin.set(plugin)
        if not plugin.is_static:
            token1 = scope_ctx.set(plugin._scope)
        try:
            super().exec_module(module)
        except RegisterNotInPluginError as e:
            log.plugin.error(f"failed to load plugin <blue>{self.plugin_id!r}</blue>:\n{e.msg}")
            plugin.dispose()
            publish(PluginLoadedFailed(self.plugin_id, e))
            raise
        except (ImportError, StaticPluginDispatchError, ReusablePluginError) as e:
            log.plugin.error(f"failed to load plugin <blue>{self.plugin_id!r}</blue>: {e.args[0]}")
            plugin.dispose()
            publish(PluginLoadedFailed(self.plugin_id, e))
            raise
        except Exception as e:
            log.plugin.exception(f"failed to load plugin <blue>{self.plugin_id!r}</blue> caused by {e!r}", exc_info=e)
            plugin.dispose()
            publish(PluginLoadedFailed(self.plugin_id, e))
            raise ImportError(f"{e!r} in {self.name!r}", name=self.name, path=self.path) from None
        finally:
            # leave plugin context
            delattr(module, "__cached__")
            if not plugin.is_static:
                scope_ctx.reset(token1)  # type: ignore
            current_plugin.reset(token)

        # get plugin metadata
        metadata: PluginMetadata | None = getattr(module, "__plugin_metadata__", None)
        if metadata and not plugin.metadata:
            plugin.metadata = metadata
        plugin._apply = getattr(module, "__plugin_apply__", None)
        if not is_sub:
            if plugin._apply:
                log.plugin.success(f"loaded plugin <blue>{self.plugin_id!r}</blue> partially applied")
            else:
                log.plugin.success(f"loaded plugin <blue>{self.plugin_id!r}</blue>")
        else:
            log.plugin.trace(f"loaded sub-plugin <r>{plugin.id!r}</r> of <y>{self.parent_plugin_id!r}</y>")
        if plugin_service.status.blocking:
            if plugin._apply:
                plugin.exec_apply()
            if plugin.config.get("$disable", False):
                plugin.disable()
            publish(Ready(), plugin._scope)
        publish(PluginLoadedSuccess(self.plugin_id))
        return


class _NamespacePath(_bootstrap_external._NamespacePath):  # type: ignore
    def _get_parent_path(self):
        parent_module_name, path_attr_name = self._find_parent_path_names()
        if parent_module_name in plugin_service.plugins:
            return plugin_service.plugins[parent_module_name].module.__path__
        return getattr(sys.modules[parent_module_name], path_attr_name)


def _path_find_spec(fullname, path=None, target=None):
    """Try to find a spec for 'fullname' on sys.path or 'path'.

    The search is based on sys.path_hooks and sys.path_importer_cache.
    """
    if path is None:
        path = sys.path
    spec = PathFinder._get_spec(fullname, path, target)  # type: ignore
    if spec is None:
        return None
    elif spec.loader is None:
        namespace_path = spec.submodule_search_locations
        if namespace_path:
            # We found at least one namespace path.  Return a spec which
            # can create the namespace package.
            spec.origin = None
            spec.submodule_search_locations = _NamespacePath(fullname, namespace_path, PathFinder._get_spec)  # type: ignore
            return spec
        else:
            return None
    else:
        return spec


class _PluginFinder(MetaPathFinder):
    @classmethod
    def find_spec(
        cls,
        fullname: str,
        path: Sequence[str] | None,
        target: ModuleType | None = None,
        origin_id_: str | None = None,
    ):
        module_spec = _path_find_spec(fullname, path, target)
        if not module_spec:
            return
        module_origin = module_spec.origin
        if not module_origin:
            if not module_spec.loader:  # namespace package
                return module_spec
            return
        if isinstance(module_spec.loader, ExtensionFileLoader):
            return
        if plug := current_plugin.get(None):
            if plug.module.__spec__ and plug.module.__spec__.origin == module_spec.origin:
                return plug.module.__spec__
            if module_spec.parent and module_spec.parent == plug.module.__name__:
                module_spec.loader = PluginLoader(fullname, module_origin, origin_id_ or fullname, plug.id)
                return module_spec
            elif module_spec.name in _SUBMODULE_WAITLIST.get(plug.module.__name__, ()):
                module_spec.loader = PluginLoader(fullname, module_origin, origin_id_ or fullname, plug.id)
                # plugin_service.referents.setdefault(module_spec.name, set()).add(plug.id)
                # _SUBMODULE_WAITLIST[plug.module.__name__].remove(module_spec.name)
                return module_spec
            if (
                module_spec.name in _ENSURE_IS_PLUGIN
                or module_spec.name.startswith("arclet.entari.builtins.")
                or module_spec.name.startswith("entari_plugin")
            ):
                module_spec.loader = PluginLoader(fullname, module_origin, origin_id_ or fullname)
                plugin_service.referents.setdefault(module_spec.name, set()).add(plug.id)
                return module_spec

        if module_spec.name in plugin_service.plugins:
            module_spec.loader = PluginLoader(fullname, module_origin, origin_id_ or fullname)
            return module_spec
        if module_spec.name in _ENSURE_IS_PLUGIN:
            module_spec.loader = PluginLoader(fullname, module_origin, origin_id_ or fullname)
            return module_spec
        if module_spec.name in plugin_service._subplugined:
            module_spec.loader = PluginLoader(
                fullname, module_origin, origin_id_ or fullname, plugin_service._subplugined[module_spec.name]
            )
            return module_spec
        if (
            any(k.startswith(module_spec.name) and k.rfind("@") != -1 for k in plugin_service.plugins)
            and (origin_id_ or fullname).rfind("@") == -1
        ):
            raise ReusablePluginError(f"reusable plugin {module_spec.name!r} cannot be imported directly")
        if module_spec.parent and module_spec.parent in plugin_service.plugins:
            module_spec.loader = PluginLoader(fullname, module_origin, origin_id_ or fullname, module_spec.parent)
            return module_spec
        if module_spec.name.rpartition(".")[0] in plugin_service.plugins:
            module_spec.loader = PluginLoader(
                fullname, module_origin, origin_id_ or fullname, module_spec.name.rpartition(".")[0]
            )
            return module_spec
        return


def find_spec(id_, package=None):
    uid_index = id_.rfind("@")
    name = id_ if uid_index == -1 else id_[:uid_index]
    fullname = resolve_name(name, package) if name.startswith(".") else name
    parent_name = fullname.rpartition(".")[0]
    if parent_name:
        parts = parent_name.split(".")
        _current = parts[0]
        if _current in plugin_service.plugins:
            parent = plugin_service.plugins[_current].module
            enter_plugin = True
        else:
            parent = __import__(_current, fromlist=["__path__"])
            enter_plugin = False
        _current += "."
        for part in parts[1:]:
            _current += part
            if _current in plugin_service.plugins:
                parent = plugin_service.plugins[_current].module
                enter_plugin = True
                _current += "."
                continue
            if _current in _ENSURE_IS_PLUGIN:
                parent = import_plugin(_current)
                if parent:
                    enter_plugin = True
                else:
                    parent = __import__(_current, fromlist=["__path__"])
                _current += "."
                continue
            if enter_plugin and (parent := import_plugin(_current)):
                pass
            else:
                enter_plugin = False
                parent = __import__(_current, fromlist=["__path__"])
            _current += "."
        try:
            parent_path = parent.__path__
        except AttributeError as e:
            raise ModuleNotFoundError(
                f"__path__ attribute not found on {parent_name!r} " f"while trying to find {fullname!r}",
                name=fullname,
            ) from e
    else:
        parent_path = None
    if isinstance(parent_path, _bootstrap_external._NamespacePath):  # type: ignore
        parent_path = _NamespacePath(parent_path._name, parent_path._path, PathFinder._get_spec)  # type: ignore
    if spec := _PluginFinder.find_spec(fullname, parent_path, origin_id_=id_):
        return spec
    module_spec = _path_find_spec(fullname, parent_path, None)
    if not module_spec:
        return
    module_origin = module_spec.origin
    if not module_origin:
        return
    if isinstance(module_spec.loader, ExtensionFileLoader):
        return
    module_spec.loader = PluginLoader(fullname, module_origin, id_)
    return module_spec


def import_plugin(id_, package=None, config: dict | None = None):
    spec = find_spec(id_, package)
    if spec:
        mod = module_from_spec(spec)
        if spec.loader:
            if isinstance(spec.loader, PluginLoader):
                spec.loader.exec_module(mod, config=config)
                protected_modules = set()
                module_name = mod.__name__
                if module_name:
                    prefix = []
                    for part in module_name.split("."):
                        prefix.append(part)
                        protected_modules.add(".".join(prefix))
                sys.modules.pop(module_name, None)
                for _imported in _IMPORTING:
                    if _imported in protected_modules or _imported in plugin_service.plugins:
                        continue
                    sys.modules.pop(_imported, None)
                _IMPORTING.clear()
            else:
                spec.loader.exec_module(mod)
        return mod
    return


sys.meta_path.insert(0, _PluginFinder())



from __future__ import annotations

import asyncio
from collections.abc import Callable
from dataclasses import dataclass, field
from pathlib import Path
import re
import sys
from types import ModuleType
from typing import Any, Generic, TypeVar, cast, overload
from weakref import finalize, proxy

from arclet.letoderea import SUBSCRIBER, Contexts, Propagator, Scope, Subscriber, define, enter_if, propagate, publish
from arclet.letoderea.breakpoint import StepOut, step_out
from arclet.letoderea.provider import Provider, ProviderFactory, TProviders
from arclet.letoderea.publisher import Publisher, _publishers
from arclet.letoderea.scope import RegisterWrapper
from creart import it
from launart import Launart, Service
from tarina import ContextModel
from tarina.tools import TCallable

from ..config import config_model_schema
from ..event.plugin import PluginLoadedFailed, PluginUnloaded
from ..exceptions import RegisterNotInPluginError, ReusablePluginError, StaticPluginDispatchError
from ..filter import parse
from ..logger import log
from .service import plugin_service

current_plugin: ContextModel[Plugin] = ContextModel("current_plugin")

T = TypeVar("T")
TS = TypeVar("TS", bound=Service)
R = TypeVar("R")


class PluginDispatcher(Generic[T]):
    def __init__(self, plugin: Plugin, event: type, name: str | None = None):
        self.publisher = define(event, name=name)
        self.plugin = plugin
        self._event = event
        self.providers: list[Provider[Any] | ProviderFactory] = []
        self.propagators: list[Propagator] = []

    # fmt: off

    def waiter(self, event: Any = None, providers: TProviders | None = None, priority: int = 15, block: bool = False) -> Callable[[Callable[..., R]], StepOut[R]]:  # noqa: E501
        if event is None:
            def decorator1(func: Callable[..., R], /) -> StepOut[R]:
                if isinstance(func, Subscriber):
                    return StepOut(func, priority, block)
                return step_out(self._event, func, providers=providers, priority=priority, block=block)
            return decorator1

        return step_out(event, providers=providers, priority=priority, block=block)

    def register(self, func: Callable[..., T] | None = None, *, priority: int = 16, providers: TProviders | None = None, once: bool = False):  # noqa: E501
        _providers = providers or []
        wrapper: RegisterWrapper[T, None] = self.plugin._scope.register(priority=priority, providers=[*self.providers, *_providers], once=once, publisher=self.publisher)  # type: ignore # noqa: E501

        old_call = RegisterWrapper.__call__

        def _call(self_, func1: Callable[..., T], /) -> Subscriber[T]:
            self.plugin.validate(func1)
            sub = old_call(self_, func1)
            sub.propagates(*self.propagators)
            RegisterWrapper.__call__ = old_call  # type: ignore
            return sub

        RegisterWrapper.__call__ = _call  # type: ignore
        if func:
            return wrapper(func)
        return wrapper

    def once(self, func: Callable[..., T] | None = None, *, priority: int = 16, providers: TProviders | None = None):  # noqa: E501
        if func:
            return self.register(func, priority=priority, providers=providers, once=True)
        return self.register(priority=priority, providers=providers, once=True)

    # fmt: on
    on = register
    handle = register

    def __call__(self, func):
        return self.register()(func)


@dataclass
class PluginMetadata:
    name: str
    """插件名称"""
    author: list[str | dict] = field(default_factory=list)
    """插件作者"""
    version: str | None = None
    """插件版本"""
    license: str | None = None
    """插件许可证"""
    urls: dict[str, str] | None = None
    """插件链接"""
    description: str | None = None
    """插件描述"""
    icon: str | None = None
    """插件图标 URL"""
    readme: str | None = None
    """插件 README"""
    classifier: list[str] = field(default_factory=list)
    """插件分类"""
    requirements: list[str] = field(default_factory=list)
    """插件依赖"""
    depend_services: list[type[Service] | str | dict] = field(default_factory=list)
    """插件依赖的服务"""
    config: Any | None = None
    """插件配置模型"""

    def get_config_schema(self) -> dict[str, Any]:
        """获取插件配置模型的 JSON Schema"""
        if self.config is None:
            return {}
        return config_model_schema(self.config)


def inject(*services: type[Service] | str | dict, _is_global: bool = False):

    async def checker(launart: Launart, ctx: Contexts):
        for service in services:
            if isinstance(service, dict):
                serv_id = service["id"] if isinstance(service["id"], str) else service["id"].id
                stage = service.get("stage", "prepared")
            else:
                serv_id = service if isinstance(service, str) else service.id
                stage = "prepared"
            sub = f"{ctx[SUBSCRIBER].callable_target.__module__}:{ctx[SUBSCRIBER].callable_target.__qualname__}"
            log.plugin.trace(f"<y>{sub}</y> waiting for service <w><u>{serv_id}</u></w> to be <g>{stage!r}</g>")
            await plugin_service.service_waiter.wait_for(serv_id)
            serv = launart.components[serv_id]
            if not serv.status.prepared:
                await serv.status.wait_for(stage)
        return True

    if _is_global:
        return enter_if(checker)

    def wrapper(func: TCallable) -> TCallable:
        try:
            plg = current_plugin.get()
        except LookupError:
            plg = func.__globals__.get("__plugin__", None)
        if not plg or not isinstance(plg, Plugin):
            raise LookupError("no plugin context found")
        plg._extra.setdefault("local_injected_services", []).append(
            (func.__qualname__, [s.id if isinstance(s, type) else s for s in services])
        )
        return propagate(enter_if(checker))(func)

    return wrapper


@dataclass
class Plugin:
    id: str
    module: ModuleType

    subplugins: set[str] = field(default_factory=set)
    config: dict[str, Any] = field(default_factory=dict)
    is_static: bool = False
    path: str = field(init=False)
    uid: str | None = None
    _metadata: PluginMetadata | None = None
    _is_disposed: bool = False
    _services: dict[str, Service] = field(init=False, default_factory=dict)
    _dispose_callbacks: list[Callable[[], None]] = field(init=False, default_factory=list)
    _config_key: str = field(init=False)
    _scope: Scope = field(init=False)
    _extra: dict[str, Any] = field(default_factory=dict, init=False)  # extra metadata for inspection
    _apply: Callable[[Plugin], Any] | None = field(default=None, init=False)

    @property
    def reusable(self) -> bool:
        return self.uid is not None and self.uid != ""

    @property
    def available(self) -> bool:
        return not self._is_disposed

    @staticmethod
    def current() -> Plugin:
        try:
            return current_plugin.get()  # type: ignore
        except LookupError:
            raise LookupError("no plugin context found") from None

    @property
    def metadata(self) -> PluginMetadata | None:
        return self._metadata

    @metadata.setter
    def metadata(self, value: PluginMetadata):
        self._metadata = value
        if value and value.depend_services:
            self._scope.propagators.append(inject(*value.depend_services, _is_global=True))  # type: ignore
            self._extra["injected_services"] = [s.id if isinstance(s, type) else s for s in value.depend_services]
        if value.readme and re.fullmatch(r"(?i:readme)\.(?i:md|markdown)", value.readme.strip()):
            readme_path = Path(self.module.__file__ or "").parent / value.readme.strip()
            try:
                value.readme = readme_path.read_text(encoding="utf-8")
            except Exception as e:
                log.plugin.error(f"failed to read readme file {readme_path}: {e!r}")
                value.readme = None
        if value.config and (
            value.config.__doc__ is None or value.config.__doc__.startswith(f"{value.config.__name__}(")
        ):
            value.config.__doc__ = value.description or value.name

    def exec_apply(self):
        if not self._apply:
            return
        log.plugin.trace(f"applying plugin <y>{self.id!r}</y>")
        token = current_plugin.set(self)
        try:
            self._apply(self)
            log.plugin.success(f"plugin <blue>{self.id!r}</blue> fully applied")
        except (ImportError, RegisterNotInPluginError, StaticPluginDispatchError, ReusablePluginError) as e:
            log.plugin.error(f"failed to load plugin <blue>{self.id!r}</blue>: {e.args[0]}")
            self.dispose()
            publish(PluginLoadedFailed(self.id, e))
            raise
        except Exception as e:
            log.plugin.exception(f"failed to load plugin <blue>{self.id!r}</blue> caused by {e!r}", exc_info=e)
            self.dispose()
            publish(PluginLoadedFailed(self.id, e))
            raise
        finally:
            current_plugin.reset(token)

    @property
    def is_available(self) -> bool:
        return self._scope.available

    def enable(self):
        if self._scope.available:
            return
        self._scope.enable()
        self.config.pop("$disable", None)
        if self.subplugins:
            subplugs = [i.removeprefix(self.id)[1:] for i in self.subplugins]
            subplugs = (subplugs[:3] + ["..."]) if len(subplugs) > 3 else subplugs
            log.plugin.trace(f"enabling sub-plugin <r>{', '.join(subplugs)}</r> of <y>{self.id}</y>")
            for subplug in self.subplugins:
                if subplug not in plugin_service.plugins:
                    continue
                plugin_service.plugins[subplug].enable()
        for ret in plugin_service.referents[self.path].copy():
            if ret not in plugin_service.plugins:
                continue
            plugin_service.plugins[ret].enable()
        if plugin_service.status.blocking:
            tasks = set()
            for serv in self._services.values():
                plugin_service.service_waiter.assign(serv.id)
                try:
                    it(Launart).add_component(serv)
                    t = asyncio.create_task(serv.status.wait_for("prepared"))
                    t.add_done_callback(tasks.discard)
                    tasks.add(t)
                except ValueError:
                    pass
            return tasks

    def disable(self):
        if not self._scope.available:
            return set()
        if self.subplugins:
            subplugs = [i.removeprefix(self.id)[1:] for i in self.subplugins]
            subplugs = (subplugs[:3] + ["..."]) if len(subplugs) > 3 else subplugs
            log.plugin.trace(f"disabling sub-plugin <r>{', '.join(subplugs)}</r> of <y>{self.id}</y>")
            for subplug in self.subplugins:
                if subplug not in plugin_service.plugins:
                    continue
                plugin_service.plugins[subplug].disable()
        for ret in plugin_service.referents[self.path].copy():
            if ret not in plugin_service.plugins:
                continue
            plugin_service.plugins[ret].disable()
        tasks = set()
        t = self._clean_service()
        t.add_done_callback(tasks.discard)
        tasks.add(t)
        self._scope.disable()
        self.config["$disable"] = True
        return tasks

    def collect(self, *disposes: Callable[[], None]):
        """收集副作用回收函数"""
        self._dispose_callbacks.extend(disposes)
        return self

    def restore(self):
        """回收所有副作用"""
        for callback in self._dispose_callbacks:
            callback()
        self._dispose_callbacks.clear()

    def __post_init__(self):
        uid_index = self.id.rfind("@")
        self.path = self.id[:uid_index] if uid_index != -1 else self.id
        self.uid = self.id[uid_index + 1 :] if uid_index != -1 else None
        self._scope = Scope.of(self.id)
        plugin_service.plugins[self.id] = self  # type: ignore
        self._config_key = self.config.pop("$path", self.id)
        allow = self.config.get("$allow", {})
        deny = self.config.get("$deny", {})
        pat = {}
        if allow:
            pat["$and"] = allow
        if deny:
            pat["$not"] = deny
        if pat:
            self._scope.propagators.append(parse(pat))
        # if self._metadata and self._metadata.depend_services:
        #     self._scope.propagators.append(inject(*self._metadata.depend_services, _is_global=True))  # type: ignore
        #     self._extra["injected_services"] = [
        #         s.id if isinstance(s, type) else s for s in self._metadata.depend_services
        #     ]
        self.is_static = self.config.pop("$static", False)
        if self.id not in plugin_service._keep_values:
            plugin_service._keep_values[self.id] = {}
        if self.path not in plugin_service.referents:
            plugin_service.referents[self.path] = set()
        if self.path not in plugin_service.references:
            plugin_service.references[self.path] = set()
        plugin_service._unloaded.discard(self.id)
        finalize(self, self.dispose, is_cleanup=True)

    def _clean_service(self):
        manager = it(Launart)

        def _gen(service):
            for serv in manager.components.values():
                if service.id in serv.required:
                    yield from _gen(serv)
            yield service

        _services = [s for serv in self._services.values() for s in _gen(serv)]

        async def _clean(services):
            if not manager.task_group:
                return
            for serv in services:
                plugin_service.service_waiter.clear(serv.id)
                if serv.id not in manager.task_group.sideload_trackers:
                    continue
                try:
                    tracker = manager.task_group.sideload_trackers[serv.id]
                    manager.remove_component(serv)
                    await asyncio.wait([tracker, asyncio.create_task(serv.status.wait_for("finished"))])
                except (ValueError, KeyError):
                    pass

        return asyncio.create_task(_clean(_services))

    def dispose(self, *, is_cleanup: bool = False):
        if not is_cleanup and self.is_static:
            return  # static plugin can only be disposed in cleanup phase
        plugin_service._unloaded.add(self.id)
        if self._is_disposed:
            return
        if not self.id.startswith(".") and self.id not in plugin_service._subplugined:
            log.plugin.debug(f"disposing plugin <y>{self.id}</y>")
        self._is_disposed = True
        tasks = set()
        t = self._clean_service()
        t.add_done_callback(tasks.discard)
        tasks.add(t)
        self._services.clear()
        if self.module.__spec__ and self.module.__spec__.cached:
            Path(self.module.__spec__.cached).unlink(missing_ok=True)
        sys.modules.pop(self.module.__name__, None)
        self.restore()
        delattr(self.module, "__plugin__")
        if self.subplugins:
            subplugs = [i.removeprefix(self.id)[1:] for i in self.subplugins]
            subplugs = (subplugs[:3] + ["..."]) if len(subplugs) > 3 else subplugs
            log.plugin.trace(f"disposing sub-plugin <r>{', '.join(subplugs)}</r> of <y>{self.id}</y>")
            for subplug in self.subplugins:
                if subplug not in plugin_service.plugins:
                    continue
                try:
                    tasks.update(plugin_service.plugins[subplug].dispose(is_cleanup=is_cleanup))
                    plugin_service._subplugined.pop(subplug, None)
                except Exception as e:
                    log.plugin.error(f"failed to dispose sub-plugin <r>{subplug}</r> caused by {e!r}")
                    plugin_service.plugins.pop(subplug, None)
            self.subplugins.clear()
        if not is_cleanup:
            publish(PluginUnloaded(self.id))
            for ref in plugin_service.references.pop(self.path):
                if ref not in plugin_service.plugins:
                    continue
                if ref in plugin_service._unloaded:
                    continue
                strip_ref = ref
                if ref.rfind("@") != -1:
                    strip_ref = ref[: ref.rfind("@")]
                if strip_ref not in plugin_service.referents:
                    continue
                if self.id not in plugin_service.referents[strip_ref]:
                    continue
                plugin_service.referents[strip_ref].remove(self.id)
                if not plugin_service.referents[strip_ref] and strip_ref not in plugin_service._direct_plugins:
                    # if no more referents, remove it
                    try:
                        tasks.update(plugin_service.plugins[ref].dispose(is_cleanup=is_cleanup))
                    except Exception as e:
                        log.plugin.error(f"failed to dispose referent plugin <r>{ref}</r> caused by {e!r}")
                        plugin_service.plugins.pop(ref, None)
            for ret in plugin_service.referents[self.path].copy():
                if ret not in plugin_service.plugins:
                    continue
                tasks.update(plugin_service.plugins[ret].disable())
        self._scope.dispose()
        self._scope.propagators.clear()
        del plugin_service.plugins[self.id]
        del self.module
        return tasks

    def dispatch(self, event, name: str | None = None):
        if self.is_static:
            raise StaticPluginDispatchError("static plugin cannot dispatch events")
        return PluginDispatcher(self, event, name=name)

    # fmt: off

    def use(self, pub: str | Publisher, func: Callable[..., Any] | None = None, *, priority: int = 16, providers: TProviders | None = None):  # noqa: E501
        if self.is_static:
            raise StaticPluginDispatchError("static plugin cannot use events by `Plugin.use`")
        if isinstance(pub, str):
            pid = pub.replace("::", "entari.event/")
        elif isinstance(pub, Publisher):
            pid = pub.id
        else:
            raise TypeError(f"invalid publisher type: {type(pub)}")
        if pid not in _publishers:
            raise LookupError(f"no publisher found: {pid}")
        disp = PluginDispatcher(self, _publishers[pid].target)
        if func:
            return disp.register(func=func, priority=priority, providers=providers)
        return disp.register(priority=priority, providers=providers)

    # fmt: on

    def validate(self, func):
        if isinstance(func, Subscriber):
            func = func.callable_target
        if func.__module__ != self.module.__name__:
            if "__plugin__" in func.__globals__ and func.__globals__["__plugin__"] is self:
                return
            raise RegisterNotInPluginError(func, self.module, self.id)

    def proxy(self):
        return proxy(self.module)

    def subproxy(self, sub_id: str):
        return proxy(plugin_service.plugins[sub_id].module)

    def service(self, serv: TS | type[TS]) -> TS:
        if self.reusable:
            raise ReusablePluginError("reusable plugin cannot provide services")
        if isinstance(serv, type):
            serv = serv()
        self._services[serv.id] = serv
        if plugin_service.status.blocking and self.is_available:
            it(Launart).add_component(serv)
            plugin_service.service_waiter.assign(serv.id)
        return serv


class RootlessPlugin(Plugin):
    # fmt: off
    @classmethod
    @overload
    def apply(cls, id: str, *, default: bool = False) -> Callable[[Callable[[RootlessPlugin], Any]], Callable[[], None]]: ...  # noqa: E501

    @classmethod
    @overload
    def apply(cls, id: str, func: Callable[[RootlessPlugin], Any], *, default: bool = False) -> Callable[[], None]: ...

    @classmethod
    def apply(cls: type[RootlessPlugin], id: str, func: Callable[[RootlessPlugin], Any] | None = None, *, default: bool = False) -> Any:  # noqa: E501
    # fmt: on
        if not id.startswith("."):
            id = f".{id}"

        def dispose():
            if id in plugin_service.plugins:
                plugin_service.plugins[id].dispose()
            else:
                plugin_service._apply.pop(id, None)

        def wrapper(func: Callable[[RootlessPlugin], Any]):
            plugin_service._apply[id] = (lambda config: cls(id, func, config), default)  # type: ignore
            return dispose

        if func:
            return wrapper(func)
        return wrapper

    def __init__(self, id: str, func: Callable[[RootlessPlugin], Any], config: dict):
        super().__init__(id, ModuleType(id), config=config)
        setattr(self.module, "__plugin__", self)
        setattr(self.module, "__file__", func.__code__.co_filename)
        self.func = func
        setattr(self.func, "__plugin__", self)
        token = current_plugin.set(self)
        try:
            func(self)
        finally:
            current_plugin.reset(token)

    def validate(self, func):
        pass


class KeepingVariable(Generic[T]):
    def __init__(self, obj: T, dispose: Callable[[T], None] | None = None):
        self.obj = obj
        self._dispose = dispose
        try:
            setattr(self.obj, "__keeping__", True)
        except AttributeError:
            pass

    def dispose(self):
        if hasattr(self.obj, "dispose"):
            self.obj.dispose()  # type: ignore
        elif self._dispose:
            self._dispose(self.obj)
        del self.obj


@overload
def keeping(id_: str, obj: T, *, dispose: Callable[[T], None] | None = None) -> T: ...


@overload
def keeping(id_: str, *, obj_factory: Callable[[], T], dispose: Callable[[T], None] | None = None) -> T: ...


# fmt: off
def keeping(id_: str, obj: T | None = None, obj_factory: Callable[[], T] | None = None, dispose: Callable[[T], None] | None = None) -> T:  # noqa: E501
# fmt: on
    if not (plug := current_plugin.get(None)):
        raise LookupError("no plugin context found")
    if id_ not in plugin_service._keep_values[plug.id]:
        if obj is None and obj_factory is None:
            raise ValueError("Either `obj` or `obj_factory` must be provided")
        _obj = obj_factory() if obj_factory else obj
        plug._extra.setdefault("kept_variables", []).append(id_)
        plugin_service._keep_values[plug.id][id_] = KeepingVariable(cast(T, _obj), dispose)  # type: ignore
    return plugin_service._keep_values[plug.id][id_].obj  # type: ignore



from arclet.alconna import Arparma

from arclet.entari import MessageChain, command, metadata
from arclet.entari.command import Match

metadata(
    "Echo",
    author=[{"name": "RF-Tar-Railt", "email": "rf_tar_railt@qq.com"}],
    description="Echo the content",
    readme="""
# Echo

该插件用于回显消息。

## 使用

- 基本用法: `echo <内容>`
- 发送转义消息: `echo -e <内容>`，例如 `echo -e @user` 会发送 `<at id=user>` 文本
- 发送反转义消息: `echo -E <内容>`，例如 `echo -E <at id=user>` 会发送 `@user` 元素
""",
    config=None,
)


@(
    command.command("echo <...content>", "显示消息")
    .option("escape", "-e|--escape # 发送转义消息")
    .option("unescape", "-E|--unescape # 发送反转义消息")
    .config(compact=True)
)
async def echo(content: Match[MessageChain], arp: Arparma):
    if arp.find("unescape"):
        return MessageChain.of(content.result.extract_plain_text())
    elif arp.find("escape"):
        return str(content.result)
    else:
        return content.result



from dataclasses import dataclass
from typing import TYPE_CHECKING

from arclet.letoderea import Contexts, Result, define, provide
from satori.client import Account
from satori.model import MessageReceipt

from ..message import MessageChain

if TYPE_CHECKING:
    from ..session import Session


@dataclass
class SendRequest:
    account: Account
    channel: str
    message: MessageChain
    session: "Session | None" = None

    def check_result(self, value) -> Result[bool | MessageChain] | None:
        if isinstance(value, bool | MessageChain):
            return Result(value)


before_send_pub = define(SendRequest, name="entari.event/before_send")
before_send_pub.bind(provide(MessageChain, call="$message"))


@before_send_pub.gather
async def req_gather(req: SendRequest, context: Contexts):
    context["$account"] = req.account
    context["$channel"] = req.channel
    context["$message"] = req.message
    if req.session:
        context["session"] = req.session


@dataclass
class SendResponse:
    account: Account
    channel: str
    message: MessageChain
    result: list[MessageReceipt]
    session: "Session | None" = None


send_pub = define(SendResponse, name="entari.event/after_send")
send_pub.bind(provide(MessageChain, call="$message"))
send_pub.bind(provide(list, target="result"))


@send_pub.gather
async def resp_gather(resp: SendResponse, context: Contexts):
    context["$account"] = resp.account
    context["$channel"] = resp.channel
    context["$message"] = resp.message
    context["result"] = resp.result
    if resp.session:
        context["$session"] = resp.session



from typing import Any

from arclet.entari.config.model import ConfigModelAction

try:
    from msgspec import Struct, convert, field, structs
    from msgspec.json import schema_components

    class MsgspecConfigAction(ConfigModelAction[Struct]):
        """Msgspec Config Model Action"""

        @classmethod
        def load(cls, data: dict[str, Any], t: type):
            return convert(data, t)

        @classmethod
        def dump(cls, obj):
            return structs.asdict(obj)

        @classmethod
        def keys(cls, obj) -> list[str]:
            return [field.name for field in structs.fields(obj)]

        @classmethod
        def schema(cls, t: type[Struct], ref_root: str = "/"):  # type: ignore
            (out,), components = schema_components((t,), ref_template=f"#{ref_root}$defs/{{name}}")
            if components:
                out["$defs"] = components
            return out

except ImportError:
    Struct = None
    field = None


def __getattr__(name):
    if name == "Struct":
        if Struct is None:
            raise ImportError("Please install `msgspec` first. Install with `pip install arclet-entari[model]`")
        return Struct
    if name == "field":
        if field is None:
            raise ImportError("Please install `msgspec` first. Install with `pip install arclet-entari[model]`")
        return field
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = ["Struct", "field"]
